/* Hdf_file.cpp */

#undef DEBUG_HDF_FILE_CPP

#ifdef DEBUG_HDF_FILE_CPP
#define PDEBUG { std::fprintf(stderr, "call to %s (%s)\n", __PRETTY_FUNCTION__, __FILE__); }
#else
#define PDEBUG
#endif

#include <cassert>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <new> // new(nothrow)
#include "Hdf_common.hpp"
#include "Hdf_file.hpp"

void
Hdf_file::Hdf_file_init(const char *filename, int32 access) {

  intn status;

  using namespace std;

  status = Hishdf(filename);
  if (status == FALSE) { 
    bad_hdf_file e(filename); 
    throw e;
  }
  
  sd_id = SDstart(filename, access); 
  if (sd_id == FAIL) {
    bad_hdf_file_opening_mode e(filename,Hdf_common::hdf_access_mode(access));
    throw e;
  }
  status = SDfileinfo(sd_id, &n_sds, &n_file_attrs);
  if (status == FAIL) {
    bad_sd_file_info e(filename);
    throw e;
  }
  
  try {
    file_attrs = new Hdf_attr[n_file_attrs];
  }
  catch (bad_alloc &e) {
    delete[] file_attrs;
    cout<<"Sorry, can't allocate the memory in "<<__FILE__<<" at "<<__LINE__<<endl;
    throw;
  }
  
  try {
    sds = new Hdf_sds[n_sds];
  }
  catch (bad_alloc &e) {
    delete[] sds;
    cout<<"Sorry, can't allocate the memory in "<<__FILE__<<" at "<<__LINE__<<endl;
    throw;
  }
  
  for (int32 iattr = 0 ; iattr < n_file_attrs ; iattr++) {
    char attr_name[MAX_NC_NAME + 1]; /* MAX_NC_NAME defined by HDF library */
    int32 attr_type;
    int32 attr_nvalues;
    
    status = SDattrinfo(sd_id, iattr, attr_name, &attr_type, &attr_nvalues);
    if (status == FAIL) {
      bad_sd_attr_info e(attr_name);
      throw e;
    }
    
    Hdf_attr hdf_attr(sd_id, iattr, attr_name, attr_type, attr_nvalues);
    file_attrs[iattr] = hdf_attr;
  }    

   
  for (int32 isds = 0 ; isds < n_sds ; isds++) {
    char sds_name[SDS_NAME_MAX_LEN];
    int32 sds_data_type;
    int32 sds_dim_sizes[SDS_MAX_RANK];
    int32 sds_nattrs;
    int32 sds_rank;
    int32 sds_id;
    
    sds_id = SDselect(sd_id, isds);
    if (sds_id == FAIL){
      bad_sd_select e(isds);
      throw e;    
    }

    status = SDgetinfo(sds_id, sds_name, &sds_rank, sds_dim_sizes, &sds_data_type, &sds_nattrs);
    if (status == FAIL) {
      bad_sd_get_info e(sds_name);
      throw e;
    }
    
    Hdf_sds hdf_sds(sds_id, sds_name, sds_rank, sds_dim_sizes, sds_data_type, sds_nattrs);
    sds[isds] = hdf_sds;
    status = SDendaccess(sds_id);
    if ( status == FAIL ) {
      bad_sd_end_access e;
      throw e;
    }
  }
  
}

Hdf_file::Hdf_file(const char *filename, int32 access, Hdf_file::sds_infos_type *sds_infos, const size_t nsds, int32 fill_mode, VOIDP fill_value, int32 compression_type, int32 compression_level)
  : filename(filename) {

  intn status;
  
  using namespace std;

  PDEBUG;

  if (access == DFACC_CREATE) {

    sd_id = SDstart(filename, access);
    if (sd_id == FAIL) {
      bad_hdf_file_opening_mode e(filename,Hdf_common::hdf_access_mode(access));
      throw e;
    }

    for (unsigned int isds = 0 ; isds < nsds ; isds++) {
      char *sds_name = sds_infos[isds].sds_name;
      int32 sds_type = sds_infos[isds].sds_type;
      int32 sds_rank = sds_infos[isds].sds_rank;
      int32 *sds_dimensions = sds_infos[isds].sds_dimensions;
      float64 cal = sds_infos[isds].sds_cal;
      float64 cal_err = sds_infos[isds].sds_cal_err;
      float64 offset = sds_infos[isds].sds_offset;
      float64 offset_err = sds_infos[isds].sds_offset_err;
      HDF_CHUNK_DEF chunk_def;
      
      int32 sds_id = SDcreate(sd_id, sds_name, sds_type, sds_rank, sds_dimensions);
      if (sds_id == FAIL) {
	bad_sd_create e(filename);
        throw e;
      }

      if (compression_type == COMP_CODE_DEFLATE) {
	for (int32 idim = 0 ; idim < sds_rank ; idim++) {
	  chunk_def.chunk_lengths[idim] = chunk_def.comp.chunk_lengths[idim] = sds_dimensions[idim];
	}
	chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	chunk_def.comp.cinfo.deflate.level = compression_level;
	
	if (SDsetchunk(sds_id, chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	  fprintf(stderr, "%s: cannot set chunks for SDS %s, compression ignored\n", filename, sds_name);
	  compression_type = COMP_CODE_NONE;
	}
	
      } else if (compression_type == COMP_CODE_SKPHUFF) {
	for (int32 idim = 0 ; idim < sds_rank ; idim++) {
	  chunk_def.chunk_lengths[idim] = chunk_def.comp.chunk_lengths[idim] = sds_dimensions[idim];
	}
	chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	chunk_def.comp.cinfo.skphuff.skp_size =  Hdf_common::hdf_sizeof(sds_type);
	
	if (SDsetchunk(sds_id, chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	  fprintf(stderr, "%s: cannot set chunks for SDS %s, compression ignored\n", filename, sds_name);
	  compression_type = COMP_CODE_NONE;
	}
      } else {
	if (compression_type != COMP_CODE_NONE) {
	  fprintf(stderr, "%s: invalid or unknown compression type, compression ignored (only Adaptive Huffman and Lempel/Ziv-77 GZIP deflation are supported today)\n", filename);
	  compression_type = COMP_CODE_NONE;
	}
      } /* if/else (compression_type) */

      status = SDsetcal(sds_id, cal, cal_err, offset, offset_err, sds_type);

      if ( fill_mode == SD_FILL) {
        if ( fill_value != NULL) { // non default fill value
            status=SDsetfillvalue(sds_id,fill_value);
            if (status == FAIL) {
                bad_sd_set_fill_value e(filename,fill_value);
                throw e;
            }
        }
        }
        else {
        status=SDsetfillmode(sds_id,SD_NOFILL);
        if (status == FAIL) {
            bad_sd_set_fill_mode e(filename,SD_NOFILL);
            throw e;
        }
      }
      
      status = SDendaccess(sds_id);
      if (status == FAIL) {
	bad_sd_end_access e;
        throw e;
      }

    } // for (isds)

    status = SDend(sd_id);
    if ( status == FAIL ) {
      bad_sd_end e;
      throw e;
    }
    access = DFACC_WRITE;
  }

  Hdf_file_init(filename, access);
}


Hdf_file::Hdf_file(const char *filename, int32 access, std::vector<sds_infos_type> sds_infos, int32 fill_mode, VOIDP fill_value, int32 compression_type, int32 compression_level)
  : filename(filename) {

  intn status;
  
  using namespace std;

  PDEBUG;

  if (access == DFACC_CREATE) {

    sd_id = SDstart(filename, access);
    if (sd_id == FAIL) {
      bad_hdf_file_opening_mode e(filename,Hdf_common::hdf_access_mode(access));
      throw e;
    }

    for (unsigned int isds = 0 ; isds < sds_infos.size() ; isds++) {
      char *sds_name = sds_infos[isds].sds_name;
      int32 sds_type = sds_infos[isds].sds_type;
      int32 sds_rank = sds_infos[isds].sds_rank;
      int32 *sds_dimensions = sds_infos[isds].sds_dimensions;
      float64 cal = sds_infos[isds].sds_cal;
      float64 cal_err = sds_infos[isds].sds_cal_err;
      float64 offset = sds_infos[isds].sds_offset;
      float64 offset_err = sds_infos[isds].sds_offset_err;
      HDF_CHUNK_DEF chunk_def;
      
      int32 sds_id = SDcreate(sd_id, sds_name, sds_type, sds_rank, sds_dimensions);
      if (sds_id == FAIL) {
	bad_sd_create e(filename);
        throw e;
      }

      if (compression_type == COMP_CODE_DEFLATE) {
	for (int32 idim = 0 ; idim < sds_rank ; idim++) {
	  chunk_def.chunk_lengths[idim] = chunk_def.comp.chunk_lengths[idim] = sds_dimensions[idim];
	}
	chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	chunk_def.comp.cinfo.deflate.level = compression_level;
	
	if (SDsetchunk(sds_id, chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	  fprintf(stderr, "%s: cannot set chunks for SDS %s\n", filename, sds_name);
	  exit(EXIT_FAILURE);
	}
	
      } else if (compression_type == COMP_CODE_SKPHUFF) {
	for (int32 idim = 0 ; idim < sds_rank ; idim++) {
	  chunk_def.chunk_lengths[idim] = chunk_def.comp.chunk_lengths[idim] = sds_dimensions[idim];
	}
	chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	chunk_def.comp.cinfo.skphuff.skp_size =  Hdf_common::hdf_sizeof(sds_type);
	
	if (SDsetchunk(sds_id, chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	  fprintf(stderr, "%s: cannot set chunks for SDS %s\n", filename, sds_name);
	  exit(EXIT_FAILURE);
	}
      } else {
	if (compression_type != COMP_CODE_NONE) {
	  fprintf(stderr, "%s: invalid or unknown compression type, compression ignored (only Adaptive Huffman and Lempel/Ziv-77 GZIP deflation are supported today)\n", filename);
	  compression_type = COMP_CODE_NONE;
	}
      } /* if/else (compression_type) */

      status = SDsetcal(sds_id, cal, cal_err, offset, offset_err, sds_type);

      if ( fill_mode == SD_FILL) {
        if ( fill_value != NULL) { // non default fill value
	  status=SDsetfillvalue(sds_id,fill_value);
	  if (status == FAIL) {
	    bad_sd_set_fill_value e(filename,fill_value);
	    throw e;
	  }
        }
      }
      else {
        status=SDsetfillmode(sds_id,SD_NOFILL);
        if (status == FAIL) {
	  bad_sd_set_fill_mode e(filename,SD_NOFILL);
	  throw e;
        }
      }
      
      status = SDendaccess(sds_id);
      if (status == FAIL) {
	bad_sd_end_access e;
        throw e;
      }
      
    } // for (isds)

    status = SDend(sd_id);
    if ( status == FAIL ) {
      bad_sd_end e;
      throw e;
    }
    access = DFACC_WRITE;
  }

  Hdf_file_init(filename, access);
}

Hdf_file::Hdf_file(const char *filename, int32 access, int32 n_sds_array, Hdf_sds *sds_array, int32 fill_mode, VOIDP fill_value, int32 compression_type, int32 compression_level)
  : filename(filename) {

  intn status;
  
  using namespace std;
  
  PDEBUG;
  
  if (access == DFACC_CREATE) {
    
    sd_id = SDstart(filename, access);
    if (sd_id == FAIL) {
      bad_hdf_file_opening_mode e(filename,Hdf_common::hdf_access_mode(access));
      throw e;
    }
    
    for (int32 isds = 0 ; isds < n_sds_array ; isds++) {
      const char *sds_name = sds_array[isds].get_name().c_str();
      int32 sds_type = sds_array[isds].get_data_type();
      int32 sds_rank = sds_array[isds].get_rank();
      int32 *sds_dimensions = new int32[sds_rank];
      sds_array[isds].get_dimensions(sds_dimensions);
      int32 sds_id = SDcreate(sd_id, sds_name, sds_type, sds_rank, sds_dimensions);
      HDF_CHUNK_DEF chunk_def;

      delete[] sds_dimensions;
      if (sds_id == FAIL) {
	bad_sd_create e(filename);
        throw e;
      }

      if (compression_type == COMP_CODE_DEFLATE) {
	for (int32 idim = 0 ; idim < sds_rank ; idim++) {
	  chunk_def.chunk_lengths[idim] = chunk_def.comp.chunk_lengths[idim] = sds_dimensions[idim];
	}
	chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	chunk_def.comp.cinfo.deflate.level = compression_level;
	
	if (SDsetchunk(sds_id, chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	  fprintf(stderr, "%s: cannot set chunks for SDS %s\n", filename, sds_name);
	  exit(EXIT_FAILURE);
	}
	
      } else if (compression_type == COMP_CODE_SKPHUFF) {
	for (int32 idim = 0 ; idim < sds_rank ; idim++) {
	  chunk_def.chunk_lengths[idim] = chunk_def.comp.chunk_lengths[idim] = sds_dimensions[idim];
	}
	chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	chunk_def.comp.cinfo.skphuff.skp_size =  Hdf_common::hdf_sizeof(sds_type);
	
	if (SDsetchunk(sds_id, chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	  fprintf(stderr, "%s: cannot set chunks for SDS %s\n", filename, sds_name);
	  exit(EXIT_FAILURE);
	}
      } else {
	if (compression_type != COMP_CODE_NONE) {
	  fprintf(stderr, "%s: invalid or unknown compression type, compression ignored (only Adaptive Huffman and Lempel/Ziv-77 GZIP deflation are supported today)\n", filename);
	  compression_type = COMP_CODE_NONE;
	}
      } /* if/else (compression_type) */

      if ( fill_mode == SD_FILL) {
        if ( fill_value != NULL) { // non default fill value
	  status=SDsetfillvalue(sds_id,fill_value);
	  if (status == FAIL) {
	    bad_sd_set_fill_value e(filename,fill_value);
	    throw e;
	  }
        }
      }
      else {
        status=SDsetfillmode(sds_id,SD_NOFILL);
        if (status == FAIL) {
	  bad_sd_set_fill_mode e(filename,SD_NOFILL);
	  throw e;
        }
      }
      
      status = SDendaccess(sds_id);
      if (status == FAIL) {
	bad_sd_end_access e;
        throw e;
      }
      
    } // for (isds)
    
    status = SDend(sd_id);
    if ( status == FAIL ) {
      bad_sd_end e;
      throw e;
    }
    access = DFACC_WRITE;
  }

  Hdf_file_init(filename, access);
  
}



Hdf_file::Hdf_file(const Hdf_file &hdf_file) :
  filename(hdf_file.filename), n_sds(hdf_file.n_sds), n_file_attrs(hdf_file.n_file_attrs)
{
  try {
    file_attrs = new Hdf_attr[n_file_attrs];
    sds = new Hdf_sds[n_sds];
    
    for (int32 iattr = 0 ; iattr < n_file_attrs ; iattr++) {
      file_attrs[iattr] = hdf_file.file_attrs[iattr];
    }
    
    for (int32 isds = 0 ; isds < n_sds ; isds++) {
      sds[isds] = hdf_file.sds[isds];
    }
  } catch (bad_alloc &e) {
    cout<<"Sorry, can't allocate the memory in "<<__FILE__<<" at "<<__LINE__<<endl;
    if ( file_attrs!=NULL ) delete[] file_attrs;
    if ( sds!=NULL ) delete[] sds;
    throw;
  }
}

Hdf_file::~Hdf_file() {

  intn status;

  PDEBUG;

  if (n_file_attrs) delete[] file_attrs;
  if (n_sds) delete[] sds;
  status = SDend(sd_id);
  if ( status == FAIL ) {
    bad_sd_end e;
    throw e;
  }
}

/* reads metadata from an sds (whose name is 'sds_name') 
 * if 'sds_data' is not NULL, loads 'sds_data' from the sds (start, stride and edges should then be given)
 * if 'sds_data' is used, it must have been allocated by the caller
 * returns an Hdf_sds object containing the sds metadata
 */
Hdf_sds
Hdf_file::read_sds(const char *sds_name, void *sds_data, int32 *start, int32 *stride, int32 *edges) {

  int32 sds_id;
  int32 sds_index;
  intn status;

  assert(sds_name != NULL);
  sds_index = SDnametoindex(sd_id, sds_name); 

  if (sds_index == FAIL) {
    bad_sds_name e(filename.c_str(),sds_name);
    throw e;
  }

  if (sds_data != NULL) {
    sds_id = SDselect(sd_id, sds_index); assert(sds_id != FAIL);
    status = SDreaddata(sds_id, start, stride, edges, (VOIDP) sds_data);
    if (status == FAIL) {
      hdf_err_code_t errcode = static_cast<hdf_err_code_t>(HEvalue(0));
      std::cerr << filename << ": " << sds_name << ": error " << errcode << ": " << HEstring(errcode) << std::endl;
      exit(EXIT_FAILURE); // FIXME: exits while debugging, will throw an exception in the final version
    }
    status = SDendaccess(sds_id);
    if ( status == FAIL ) {
      bad_sd_end_access e;
      throw e;
    }
  }

  return sds[sds_index];

}

/* same as above, where sds_name is replaced by sds_index
 */
Hdf_sds
Hdf_file::read_sds(const int32 sds_index, void *sds_data, int32 *start, int32 *stride, int32 *edges) {

  int32 sds_id;
  intn status;

  if (sds_data != NULL) {
    sds_id = SDselect(sd_id, sds_index); assert(sds_id != FAIL);
    status = SDreaddata(sds_id, start, stride, edges, (VOIDP) sds_data); 
    if (status == FAIL) {
      hdf_err_code_t errcode = static_cast<hdf_err_code_t>(HEvalue(0));
      std::cerr << filename << ": index " << sds_index << ": error " << errcode << ": " << HEstring(errcode) << std::endl;
      exit(EXIT_FAILURE); // FIXME: exits while debugging, will throw an exception in the final version
    }
    status = SDendaccess(sds_id);
    if ( status == FAIL ) {
      bad_sd_end_access e;
      throw e;
    }
  }

  return sds[sds_index];

}



/* stores 'sds_data' in the sds whose name is 'sds_name'
 * start, stride and edges must be given
 */
void
Hdf_file::write_sds(const char *sds_name, const void *sds_data, int32 *start, int32 *stride, int32 *edges, float64 cal, float64 offset, float64 cal_err, float64 off_err) {

  int32 sds_id;
  int32 sds_index;
  intn status;

  assert(sds_name != NULL);
  sds_index = SDnametoindex(sd_id, sds_name);

  if (sds_index == FAIL) {
    bad_sds_name e(filename.c_str(),sds_name);
    throw e;
  }

  if (sds_data != NULL) {
    sds_id = SDselect(sd_id, sds_index); assert(sds_id != FAIL);
    status = SDwritedata(sds_id, start, stride, edges, (VOIDP) sds_data);
    if (status == FAIL) {
      hdf_err_code_t errcode = static_cast<hdf_err_code_t>(HEvalue(0));
      std::cerr << filename << ": " << sds_name << ": error " << errcode << ": " << HEstring(errcode) << std::endl;
      exit(EXIT_FAILURE); // FIXME: exits while debugging, will throw an exception in the final version
    }

    if (cal != 0) {
      int32 data_type;
      char sds_name_2[256]; /* not used */
      int32 rank; /* not used */
      int32 dim_sizes[32]; /* not used */
      int32 n_attrs; /* not used */

      status = SDgetinfo(sds_id, sds_name_2, &rank, dim_sizes, &data_type, &n_attrs); assert(status != FAIL);
      status = SDsetcal(sds_id, cal, cal_err, offset, off_err, data_type); assert(status != FAIL);
    }

    status = SDendaccess(sds_id);
    if ( status == FAIL ) {
      bad_sd_end_access e;
      throw e;
    }
  }
}

// same as above, where sds_name is replaced by sds_index
void
Hdf_file::write_sds(const int32 sds_index, const void *sds_data, int32 *start, int32 *stride, int32 *edges, float64 cal, float64 offset, float64 cal_err, float64 off_err) {

  int32 sds_id;
  intn status;

  if (sds_data != NULL) {
    sds_id = SDselect(sd_id, sds_index); assert(sds_id != FAIL);
    status = SDwritedata(sds_id, start, stride, edges, (VOIDP) sds_data);
    if (status == FAIL) {
      hdf_err_code_t errcode = static_cast<hdf_err_code_t>(HEvalue(0));
      std::cerr << filename << ": index " << sds_index << ": error " << errcode << ": " << HEstring(errcode) << std::endl;
      exit(EXIT_FAILURE); // FIXME: exits while debugging, will throw an exception in the final version
    }

    if (cal != 0) {
      int32 data_type;

      status = SDgetinfo(sds_id, NULL, NULL, NULL, &data_type, NULL); assert(status != FAIL);
      status = SDsetcal(sds_id, cal, cal_err, offset, off_err, data_type); assert(status != FAIL);
    }

    status = SDendaccess(sds_id);
    if ( status == FAIL ) {
      bad_sd_end_access e;
      throw e;
    }
  }
}

Hdf_sds
Hdf_file::get_sds(const int32 &sds_index) const {
  assert(sds_index < n_sds);
  return sds[sds_index];
}

Hdf_sds
Hdf_file::get_sds(const char *sds_name) const {

  int32 sds_index;

  assert(sds_name != NULL);
  sds_index = SDnametoindex(sd_id, sds_name);

  if (sds_index == FAIL) {
    bad_sds_name e(filename.c_str(),sds_name);
    throw e;
  }

  return get_sds(sds_index);

}

Hdf_attr
Hdf_file::get_attr(const char *file_attr) const {
  
  int32 attr_index;

  assert(file_attr != NULL);
  attr_index = SDfindattr(sd_id, file_attr);

  if (attr_index == FAIL) { /* FIXME: should throw an exception if file_attr is not found */
    fprintf(stderr, "%s: file attribute not found in %s\n", file_attr, filename.c_str());
    exit(EXIT_FAILURE);
  }

  assert(attr_index < n_file_attrs);
  return file_attrs[attr_index];
  
}

void
Hdf_file::get_sds_array(Hdf_sds *sds_array) {

  assert(sds_array);

  for (int32 isds = 0 ; isds < n_sds ; isds++) {
    sds_array[isds] = sds[isds];
  }

}

void
Hdf_file::set_sds(int32 sds_index, Hdf_sds sds) {

  assert(sds_index > 0);
  assert(sds_index < n_sds);

  this->sds[sds_index] = sds;

}


/* returns a string representation of the file attributes list */
std::string
Hdf_file::list_of_attributes(const int32 chars_to_display, const int32 numbers_to_display) const {

  std::string s;

  Hdf_attr::chars_to_display = chars_to_display;
  Hdf_attr::numbers_to_display = numbers_to_display;

  s = "";

  for (int32 iattr = 0 ; iattr < n_file_attrs ; iattr++) {
    s += file_attrs[iattr].to_string();
    s += '\n';
  }

  return s;
  
}

/* returns a string representation of the list of sds in the file */
std::string
Hdf_file::list_of_sds() const {

  std::string s;

  s = "";
  
  for (int32 isds = 0 ; isds < n_sds ; isds++) {
    s += sds[isds].to_string();
    s += '\n';
  }

  return s;

}

Hdf_file &
Hdf_file::operator= (const Hdf_file &hdf_file) {

  if (&hdf_file == this) return *this; // protect against self assignation

  this->filename = hdf_file.filename;
  this->n_sds = hdf_file.n_sds;
  this->n_file_attrs = hdf_file.n_file_attrs;

  if (n_file_attrs) delete[] file_attrs;
  if (n_sds) delete[] sds;

  file_attrs = new(std::nothrow) Hdf_attr[n_file_attrs];  assert(file_attrs != NULL); /* FIXME: nothrow to be replaced by try ... catch */
  sds = new(std::nothrow) Hdf_sds[n_sds];  assert(sds != NULL); /* FIXME: nothrow to be replaced by try ... catch */

  for (int32 iattr = 0 ; iattr < n_file_attrs ; iattr++) {
    file_attrs[iattr] = hdf_file.file_attrs[iattr];
  }

  for (int32 isds = 0 ; isds < n_sds ; isds++) {
    sds[isds] = hdf_file.sds[isds];
  }

  return *this;
}

void Hdf_file::set_file_attr( const char * attr_name, const int val_type, const int32 val_length, VOIDP attr_val )
{
  intn status;
  status=SDsetattr (sd_id, attr_name, val_type, val_length, attr_val);
  if (sd_id == FAIL) {
    bad_set_file_attribute e(filename.c_str(),attr_name);
    throw e;
  }
}

void Hdf_file::set_sds_attr( const int32 sds_index, const char * attr_name, const int val_type, const int32 val_length, VOIDP attr_val )
{
  intn status;
  status=SDsetattr(sds_index, attr_name, val_type, val_length, attr_val);
  if (sd_id == FAIL) {
    std::string s=get_sds_name(sds_index);
    bad_set_sds_attribute e(filename.c_str(),s.c_str(),attr_name);
    throw e;
  }
}

void Hdf_file::set_sds_attr( const char * sds_name, const char * attr_name, const int val_type, const int32 val_length, VOIDP attr_val )
{
  int32 sds_index = get_sds_index(sds_name);
  set_sds_attr(sds_index,attr_name,val_type,val_length,attr_val);
}

const int32 Hdf_file::get_sds_index( const char * sds_name )
{
  assert(sds_name != NULL);

  int32 sds_index = SDnametoindex(sd_id, sds_name);
  if (sds_index == FAIL) {
    bad_sds_name e(filename.c_str(),sds_name);
    throw e;
  }
  return sds_index;
}

std::string Hdf_file::get_sds_name( const int32 & sds_index )
{
  Hdf_sds sds = get_sds(sds_index);
  return sds.get_name();
}
