#!/usr/bin/perl

use Time::Local;
use File::Basename;
use Getopt::Std;

# in this script, we'll call 'date' a string in the format 'YYYYMMDDHHMM(SS)'
# 'time' will be a number of seconds since the 1st jan 1970

sub parse_date($) {
	my $date = shift(@_);
	my $second = substr($date, -2, 2);
	my $minute = substr($date, -4, 2);
	my $hour   = substr($date, -6, 2);
	my $day    = substr($date, -8, 2);
	my $month  = substr($date, -10, 2);
	my $year   = substr($date, -14, 4);
	
	return ($second, $minute, $hour, $day, $month, $year);
}

sub date_to_time($) {
	my $date = shift(@_);
	my ($second, $minute, $hour, $day, $month, $year) = parse_date($date);
	my $time = timegm($second + 0, $minute + 0, $hour + 0, $day + 0, $month - 1, $year - 1900);
	return $time;
}

sub time_to_date($) {
	my $time = shift(@_);
	my ($second, $minute, $hour, $day, $month, $year) = gmtime($time);
	$month++;
	$year += 1900;
	$second = sprintf "%2.2d", $second; # unused
	$minute = sprintf "%2.2d", $minute;
	$hour   = sprintf "%2.2d", $hour;
	$day    = sprintf "%2.2d", $day;
	$month  = sprintf "%2.2d", $month;
	$year   = sprintf "%2.2d", $year;
	my $date = "$year$month$day$hour$minute$second";
	
	return $date;
}

sub time_to_date_in_minutes($) {
	my $time = shift(@_);
	my ($second, $minute, $hour, $day, $month, $year) = gmtime($time);
	$month++;
	$year += 1900;
	$second = sprintf "%2.2d", $second; # unused
	$minute = sprintf "%2.2d", $minute;
	$hour   = sprintf "%2.2d", $hour;
	$day    = sprintf "%2.2d", $day;
	$month  = sprintf "%2.2d", $month;
	$year   = sprintf "%2.2d", $year;
	my $date = "$year$month$day$hour$minute";
	
	return $date;
}

die "$0: environment variable COOKIE_CUTTER_ROOT not defined\n" unless ($ENV{COOKIE_CUTTER_ROOT});
my $COOKIE_CUTTER_ROOT=$ENV{COOKIE_CUTTER_ROOT};

my $orbital_inclination='98.2';
my $orbital_period='98.88';
my $extraction_width='10';
getopt('i:T:w:');

$orbital_inclination = $opt_i if ($opt_i);
$orbital_period = $opt_T if ($opt_T);
$extraction_width = $opt_w if ($opt_w);
$opt_I = '-I' if ($opt_I);

if (@ARGV < 3) {
	print "usage: $0 [OPTIONS] <node_file> <input_directory> <output_file>\n\n";
	print "  node_file: a text file with three columns: YYYY-MM-DD hh:mm:ss Node_Longitude\n";
	print "  (warning: Longitude is expected in degrees, between 0 (East of Greenwich) and 360 (West of Greenwich)\n";
	print "  input_directory: must contain at most one day of XRIT files (name H-000-MSG1__-MSG1________-?????____-CYCLE____-????????????)\n\n";
	print "OPTIONS:\n";
	print "  -i <orbital_inclination()>\n";
	print "  -T <orbital_period(minutes)>\n";
	print "  -w <extraction_width()>\n";
	print "  -k                              keeps temporary files (for debugging purposes)\n";
	exit 1;
}

my $BINDIR="$COOKIE_CUTTER_ROOT/bin";
my $NODE_FILE=$ARGV[0];
my $INPUT_DIR=$ARGV[1];
my ($OUTPUT_FILE, $OUTPUT_DIR)=fileparse($ARGV[2]);
open NODE_FILE, $NODE_FILE or die "$0: $NODE_FILE: $!";

if (! -d $OUTPUT_DIR ) {
	mkdir $OUTPUT_DIR or die "$0: $OUTPUT_DIR: $!";
}

while (my $input_line = <NODE_FILE>) { # processes each entry of the node file
	my ($node_date_, $node_hour_, $node_lon_) = split /\s+/, $input_line;
	
	# original longitude convention is : [0,360]
	# change it to : [-180,+180]
	$node_lon_ -= 360. if ($node_lon_ > 180.);
	
	next if ($node_lon_ <= -90. || $node_lon_ >= 90.);
	my $node_lon = sprintf "%+06.2f", $node_lon_;
	
	my ($year, $month, $day) = split /-/, $node_date_;
	my ($hour, $minute, $second) = split /:/, $node_hour_;
	my $node_date="$year$month$day$hour$minute$second";
	
	# node_quarter_time is the same as node_time, with a quarter of hour precision
	# sample: 20050601131603 becomes 20050601131500

	my $node_quarter_time = date_to_time($node_date);
	$node_quarter_time = sprintf('%d', $node_quarter_time/900.)*900; # 900 s in a quarter of hour
	
	# does the extraction between $first_quarter_time (one half of an hour before $node_quarter_time)
	# and $last_quarter_time (one half of an hour after $node_quarter_time)
	my $first_quarter_time = $node_quarter_time - 2*900;
	my $last_quarter_time = $node_quarter_time + 2*900;
	
	for (my $quarter_time = $first_quarter_time ; $quarter_time <= $last_quarter_time ; $quarter_time += 900) {
		my $quarter_date_in_minutes = time_to_date_in_minutes($quarter_time);
		my $next_quarter_date_in_minutes = time_to_date_in_minutes($quarter_time + 900);
		my $quarter_mean_time = $quarter_time + 450;
		my $quarter_mean_date = time_to_date($quarter_mean_time);
		my $extraction_file = "ICF-IVW_${quarter_date_in_minutes}_lon=${node_lon}.hdf";
		my $cmd="$BINDIR/msg2orbit.pl -o $OUTPUT_DIR/$extraction_file $opt_I -w $extraction_width -l $node_lon -M $quarter_mean_date -N $node_date -i $orbital_inclination -T $orbital_period $INPUT_DIR $quarter_date_in_minutes";
		#print "$cmd\n" if ($opt_v);
		print `$cmd`;
		if (-f "$OUTPUT_DIR/$extraction_file") {
			my $cmd="$BINDIR/cutslot $OUTPUT_DIR/$extraction_file -o $OUTPUT_DIR/ICF-IVW_${quarter_date_in_minutes}-${next_quarter_date_in_minutes}_lon=${node_lon}.hdf";
			print "$cmd\n" if ($opt_v);
			print `$cmd`;
		}
	}
}

close NODE_FILE;

# generation of HDF mosaic
my $cmd="$BINDIR/msg_mosaic -c gzip -o $OUTPUT_DIR/$OUTPUT_FILE.hdf $OUTPUT_DIR/ICF-IVW_????????????-????????????_lon=[+-]??.??.hdf";
#print "$cmd\n" if ($opt_v);
print `$cmd`;

# generation of browses (images)
$cmd="$BINDIR/sds2ppm -R=7,50,600 -G=6,50,600 -B=5,50,600 -int=0,0,40,90,120,190,255,255 $OUTPUT_DIR/$OUTPUT_FILE.hdf -of=$OUTPUT_DIR/$OUTPUT_FILE.ppm";
print "$cmd\n" if ($opt_v);
print `$cmd`;

# removing of temporary files
unless ($opt_k) {
   $cmd="rm -f $OUTPUT_DIR/ICF-IVW_*.hdf";
   print "$cmd\n" if ($opt_v);
   print `$cmd`;
}

exit 0;
