/* cutslot.cpp */

#include <cstdlib>
#include <ctime>
#include <cfloat>
#include <iostream>
#include <unistd.h>
#include "Hdf/Hdf.hpp"

#define DEBUG

using namespace std;

#define APPNAME "cutslot"
#define STRING_MAXLEN 255
#define DEFAULT_OUTPUT_FILENAME "nearest.hdf"
#define OPT_VERBOSE 0x1

char output_filename[STRING_MAXLEN];
int flags;

/* some inline help */
void usage() {
  cerr << "usage: " APPNAME " [OPTIONS] slot1 [slot2 slot3...]" << endl << endl;
  cerr << "OPTIONS: " << endl;
  cerr << "  -o <output_filename>                        output file (default is " DEFAULT_OUTPUT_FILENAME ")" << endl;
  cerr << "  -v                                          verbose" << endl;
  cerr << endl;
  exit (EXIT_FAILURE);
}

/* options */
void parse_options(int *argc, char **argv[]) {
  int option;
  const char *options_list = "o:v";
  
  flags = 0;
  
  while ((option = getopt(*argc, *argv, options_list)) != -1 ) {
    switch(option) {
    case 'o' :
      /* specifies output filename */
      strncpy(output_filename, optarg, STRING_MAXLEN);
      break;
    case 'v' :
      /* enables verbose mode */
      flags |= OPT_VERBOSE;
      break;
    case '?' :
      fprintf(stderr,"option %c : unknown\n\n",optopt);
      usage();
    default :
      fprintf(stderr, APPNAME ": parse options : fatal error\n");
      exit(8);
    } /* switch */
  } /* while */
  *argc -= optind - 1;
  *argv += optind - 1;
}

template <typename T>
void load_dataset(/* inputs  */ const char *filename, const char *dataset, 
		  /* outputs */ int32 *data_type, int32 *rank, int32 *&dimensions /* delete[] it yourself */, T *&buffer /* delete[] it yourself */,
				float64 *slope = NULL, float64 *offset = NULL) {

  int32 data_type_;
  int32 rank_;
  int32* start;
  size_t bufsiz;
  float64 slope_;
  float64 offset_;

  assert(filename);
  assert(dataset);

  Hdf_file hdf_file(filename, DFACC_READ);
  Hdf_sds hdf_sds = hdf_file.get_sds(dataset);
  rank_ = hdf_sds.get_rank();
  if (rank) *rank = rank_;
  data_type_ = hdf_sds.get_data_type();
  if (data_type) *data_type = data_type_;
  dimensions = new int32[rank_];
  start = new int32[rank_];
  hdf_sds.get_dimensions(dimensions);

  if (slope || offset) {
    try {
      Hdf_attr attr_slope = hdf_sds.get_attribute("scale_factor");
      Hdf_attr attr_offset = hdf_sds.get_attribute("add_offset");
      attr_slope.get_value(&slope_);
      attr_offset.get_value(&offset_);
      if (flags & OPT_VERBOSE) cout << "  slope: " << slope_ << "  offset: " << offset_ << endl;
      if (slope) *slope = slope_;
      if (offset) *offset = offset_;
    }
    catch (string err_msg) {
      /* do nothing (some datasets may not have calibration coefficients) */
    }
  }

  assert(Hdf_common::hdf_sizeof(data_type_) == sizeof(T));

  bufsiz = sizeof(T);
  for (int idim = 0 ; idim < rank_ ; idim++) {
    start[idim] = 0;
    bufsiz *= dimensions[idim];
  }
  buffer = new T[bufsiz];
  hdf_sds.read(buffer, start, NULL, dimensions);
  delete[] start;
  
}

/* finds for each pixel of three SEVIRI contiguous slots, which one is the nearest in time
 * from the Moving Satellite passage 
 */
void find_nearest_slots(/* inputs */ int nslots, char **slots, 
			/* outputs */ int &nrows, int &ncols, int* &nearest_islots /* delete[] it yourself */) {
  uint32 *moving_sat_times;
  uint32 **seviri_acq_times; // seviri acquisition time from all the slots
  time_t *seviri_acq_time; // nearest seviri acquisition time at pixel scale (from the Moving Satellite Passage)
  int32 *dtime;
  int32 data_type;
  int32 rank;
  int32 *dimensions;

  seviri_acq_times = new (uint32*)[nslots];
  seviri_acq_time  = new time_t[nslots];
  dtime = new int32[nslots];

  load_dataset(slots[0], "Moving_Sat_Time", &data_type, &rank, dimensions, moving_sat_times);
  nrows = dimensions[0];
  ncols = dimensions[1];

  for (int islot = 0 ; islot < nslots ; islot++) {
    load_dataset(slots[islot], "Seviri_Acq_Time", &data_type, &rank, dimensions, seviri_acq_times[islot]);
    assert(nrows == dimensions[0]);
    assert(ncols == dimensions[1]);
  }

  delete[] dimensions;

  nearest_islots = new int[nrows*ncols];

  for (int irow = 0 ; irow < nrows ; irow++) {
    for (int icol = 0 ; icol < ncols ; icol++) {
      const time_t moving_sat_time = moving_sat_times[irow*ncols + icol];
      int32 dtime_min;
      
      if (moving_sat_time == -1) {
	nearest_islots[irow*ncols + icol] = -1;
	continue;
      }

      for (int islot = 0 ; islot < nslots ; islot++) {
	seviri_acq_time[islot] = seviri_acq_times[islot][irow*ncols + icol];
	dtime[islot] = abs(moving_sat_time - seviri_acq_time[islot]);
      }

      dtime_min = dtime[0];
      nearest_islots[irow*ncols + icol] = 0;
      for (int islot = 1 ; islot < nslots ; islot++) {
	if (dtime[islot] < dtime_min) {
	  dtime_min = dtime[islot];
	  nearest_islots[irow*ncols + icol] = islot;
	}
      }
      if (dtime_min > 450) nearest_islots[irow*ncols + icol] = -1; // Moving Satellite should always be at most at 7'30 from SEVIRI (SEVIRI slot half duration)
    } // for icol
  } // for irow

  delete[] moving_sat_times;
  delete[] seviri_acq_time;
  delete[] dtime;
  for (int islot = 0 ; islot < nslots ; islot++) delete[] seviri_acq_times[islot];
  delete[] seviri_acq_times;

}

template <typename T>
void fill_dataset(Hdf_file &hdf_file, const char *sds_name, T *buffer, size_t bufsiz) {
  
  int32 *dimensions;
  int32 *start;
  int32 rank;
  int32 data_type;
  size_t bufsiz_;
  int nrows;
  int ncols;

  assert(buffer);
  assert(sds_name);
  Hdf_sds hdf_sds = hdf_file.get_sds(sds_name);
  rank = hdf_sds.get_rank();
  assert(rank == 2);

  data_type = hdf_sds.get_data_type();
  dimensions = new int32[rank];
  start = new int32[rank];
  hdf_sds.get_dimensions(dimensions);
  nrows = dimensions[0];
  ncols = dimensions[1];
  
  bufsiz_ = sizeof(T)*nrows*ncols;
  assert(bufsiz == bufsiz_);

  for (int idim = 0 ; idim < rank ; idim++) {
    start[idim] = 0;
  }
  
  hdf_file.write_sds(sds_name, buffer, start, NULL, dimensions);

  delete[] dimensions;
  delete[] start;
}

template <typename T>
void fill_dataset(Hdf_file &hdf_file, const char *sds_name, T *buffer, size_t bufsiz, float64 slope, float64 offset) {
  
  int32 *dimensions;
  int32 *start;
  int32 rank;
  int32 data_type;
  size_t bufsiz_;
  int nrows;
  int ncols;

  assert(buffer);
  assert(sds_name);
  Hdf_sds hdf_sds = hdf_file.get_sds(sds_name);
  rank = hdf_sds.get_rank();
  assert(rank == 2);

  data_type = hdf_sds.get_data_type();
  dimensions = new int32[rank];
  start = new int32[rank];
  hdf_sds.get_dimensions(dimensions);
  nrows = dimensions[0];
  ncols = dimensions[1];
  
  bufsiz_ = sizeof(T)*nrows*ncols;
  assert(bufsiz == bufsiz_);

  for (int idim = 0 ; idim < rank ; idim++) {
    start[idim] = 0;
  }
  
  hdf_file.write_sds(sds_name, buffer, start, NULL, dimensions, slope, offset);

  delete[] dimensions;
  delete[] start;
}


template <typename T>
void build_dataset_of_nearest_pixels(int nslots, char **slots, const int *nearest_islots, Hdf_file &hdf_output_file, const char *sds_name, const T undef_value) {
  int32 data_type;
  int32 rank;
  int32 *dimensions;
  int nrows;
  int ncols;
  T **buffer;
  T *nearest;
  size_t *buffer_siz;
  size_t nearest_siz;
  float64 *slope;
  float64 *offset;
  float64 slope_min = DBL_MAX; // minimal slope for all slopes
  float64 offset_min = 0.; // offset corresponding to a minimal slope
  int islot_min = -1; // slot index for which slope is minimal

  if (flags & OPT_VERBOSE) cout << APPNAME ": stores nearest(" << sds_name << ") into " << hdf_output_file.get_filename() << endl;

  buffer = new (T*)[nslots]; assert(buffer);
  buffer_siz = new size_t[nslots]; assert(buffer_siz);
  slope = new float64[nslots]; assert(slope);
  offset = new float64[nslots]; assert(offset);

  for (int i = 0 ; i < nslots ; i++) {
    slope[i] = 0.;
    offset[i] = 0.;
    load_dataset(slots[i], sds_name, &data_type, &rank, dimensions, buffer[i], &slope[i], &offset[i]);
    assert(rank == 2);
    nrows = dimensions[0];
    ncols = dimensions[1];
    buffer_siz[i] = Hdf_common::hdf_sizeof(data_type)*nrows*ncols;

    if (slope[i] < slope_min) {
      slope_min = slope[i];
      offset_min = offset[i];
      islot_min = i;
    }
  }

  assert(slope_min != DBL_MAX);

  for (int i = 0 ; i < nslots ; i++) {
    assert(buffer_siz[0] == buffer_siz[i]);
  }

  nearest_siz = sizeof(T)*nrows*ncols;
  assert(nearest_siz == buffer_siz[0]);

  nearest = new T[nrows*ncols];
  
  for (int irow = 0 ; irow < nrows ; irow++) {
    for (int icol = 0 ; icol < ncols ; icol++) {
      int nearest_islot = nearest_islots[irow*ncols + icol];
      if (nearest_islot >= 0) {
	nearest[irow*ncols + icol] = buffer[nearest_islot][irow*ncols + icol];
      }
      else  {
	nearest[irow*ncols + icol] = undef_value;
      }
    }
  }
  if (slope_min > 0.) fill_dataset(hdf_output_file, sds_name, nearest, nearest_siz, slope_min, offset_min);
  else fill_dataset(hdf_output_file, sds_name, nearest, nearest_siz);

  for (int i = 0 ; i <  nslots ; i++) {
    delete[] buffer[i];
  }

  delete[] buffer;
  delete[] buffer_siz;
  delete[] slope;
  delete[] offset;

  delete[] nearest;
}


void build_output_file(/* inputs */ int nslots, char **slots, const int *nearest_islots,
		       /* outputs */ const char *output_filename) {
    
  int32 n_sds;
  Hdf_sds *sds_array;

  Hdf_file hdf_input_file(slots[0], DFACC_READ);
  n_sds = hdf_input_file.get_n_sds();
  sds_array = new Hdf_sds[n_sds];
  hdf_input_file.get_sds_array(sds_array);
  Hdf_file hdf_output_file(output_filename, DFACC_CREATE, n_sds, sds_array);

  for (int isds = 0 ; isds < n_sds ; isds++) {
    const char *sds_name = sds_array[isds].get_name().c_str();
    int32 sds_type = sds_array[isds].get_data_type();

    switch (sds_type) {
    case (DFNT_UINT8) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<uint8>(-1));
      break;
    case (DFNT_INT8) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<int8>(-1));
      break;
    case (DFNT_UINT16) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<uint16>(-1));
      break;
    case (DFNT_INT16) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<int16>(-1));
      break;
    case (DFNT_UINT32) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<uint32>(-1));
      break;
    case (DFNT_INT32) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<int32>(-1));
      break;
    case (DFNT_FLOAT32) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<float32>(0.));
      break;
    case (DFNT_FLOAT64) : 
      build_dataset_of_nearest_pixels(nslots, slots, nearest_islots, hdf_output_file, sds_name, static_cast<float64>(0.));
      break;
      
    default : 
      cerr << APPNAME ": unexpected or not yet implemented hdf type (hdf type code = " << sds_type << ")" << endl;
      exit (EXIT_FAILURE);
    } // switch

  }

  delete[] sds_array;
}


int main(int argc, char *argv[]) {

  char **input_slots;
  int *nearest_slots;
  int nslots;
  int nrows;
  int ncols;
  
  parse_options(&argc, &argv);

  if (argc < 2) usage();

  if (! *output_filename) {
    strncpy(output_filename, DEFAULT_OUTPUT_FILENAME, STRING_MAXLEN);
  }

  nslots = argc - 1;
  input_slots = new (char *)[nslots];
  for (int islot = 0 ; islot < nslots ; islot++) {
    input_slots[islot] = argv[islot + 1];
  }
  find_nearest_slots(nslots, input_slots, nrows, ncols, nearest_slots);
  build_output_file(nslots, input_slots, nearest_slots, output_filename);
  delete[] nearest_slots;
  delete[] input_slots;
  exit(EXIT_SUCCESS);
}
