/* geostat.c */

#include <string.h> /* strlen */
#include <math.h> /* lround */
#include "geostat.h"
#include "fortran.h"
#include "msg_grid.h"

/* lround prototype is not included with math.h 
 * (maybe a bug in Linux implementation of math.h ?)
 */
extern long int lround(double x);

/* C interface for Fortran code */

#ifdef __cplusplus
# define C_LINK "C"
#else
# define C_LINK
#endif

extern C_LINK void FORTRAN_NAME(geolllp)(double *, double *, double *, double *, int *, int *);
extern C_LINK void FORTRAN_NAME(geolpll)(double *, double *, double *, double *, int *, int *);
extern C_LINK void FORTRAN_NAME(defgeo)(const char *, int *, int);
extern C_LINK void FORTRAN_NAME(geovis)(double *, double *, int *, double *, double *, int *);

geostat_err_t geostat_init(const char *satellite)
{
	int errcode;
	
	FORTRAN_NAME(defgeo)(satellite, &errcode, strlen(satellite));
	if (errcode) return GEOSTAT_ERR_BAD_INIT;
	return GEOSTAT_ERR_OK;
}

geostat_err_t geostat_latlon_to_lincol_double(
	double latitude,
	double longitude,
	double *line,
	double *column)
{
	int errcode;
	const channel_t channel = 1; /* always 1 for MSG low resolution channels (3 km x 3km at subsatellite point) */

	FORTRAN_NAME(geolllp)((double *) &latitude, (double *) &longitude, line, column, (int *) &channel, &errcode);
	
	/* flips line and column returned by geolllp (output of geolllp gives line 1 at south and column 1 at east */
	*line = MSG_NLINES - *line + 1;
	*column = MSG_NCOLUMNS - *column + 1;
	
	return (geostat_err_t) errcode;

}

geostat_err_t geostat_lincol_to_latlon_double(
	double line,
	double column,
	double *latitude,
	double *longitude)
{
	int errcode;
	const channel_t channel = 1; /* always 1 for MSG low resolution channels (3 km x 3km at subsatellite point) */

	/* flips line and column given geolllp (input of geolllp expects line 1 at south and column 1 at east */
	line = MSG_NLINES - line + 1;
	column = MSG_NCOLUMNS - column + 1;
	
	FORTRAN_NAME(geolpll)(&line, &column, latitude, longitude, (int *) &channel, &errcode);
	
	return (geostat_err_t) errcode;

}


geostat_err_t geostat_latlon_to_lincol(
	const double latitude,
	const double longitude,
	unsigned long *line,
	unsigned long *column)
{
	int errcode;
	const channel_t channel = 1; /* always 1 for MSG low resolution channels (3 km x 3km at subsatellite point) */
	double dbl_line, dbl_column;
	
	FORTRAN_NAME(geolllp)((double *) &latitude, (double *) &longitude, &dbl_line, &dbl_column, (int *) &channel, &errcode);
	
	/* flips line and column returned by geolllp (output of geolllp gives line 1 at south and column 1 at east */
	*line = MSG_NLINES - lround(dbl_line) + 1;
	*column = MSG_NCOLUMNS - lround(dbl_column) + 1;
	
	return (geostat_err_t) errcode;

}


geostat_err_t geostat_lincol_to_latlon(
	const unsigned long line,
	const unsigned long column,
	double *latitude,
	double *longitude)
{
	int errcode;
	const channel_t channel = 1; /* always 1 for MSG low resolution channels (3 km x 3km at subsatellite point) */

	/* flips line and column given geolllp (input of geolllp expects line 1 at south and column 1 at east */
	double dbl_line = (double) MSG_NLINES - line + 1;
	double dbl_column = (double) MSG_NCOLUMNS - column + 1;
	
	FORTRAN_NAME(geolpll)((double *) &dbl_line, (double *) &dbl_column, latitude, longitude, (int *) &channel, &errcode);
	
	return (geostat_err_t) errcode;

}

geostat_err_t geostat_get_azimut_and_zenithal(
	const double latitude,
	const double longitude,
	double *azimut,
	double *zenithal)
{
	const unsigned int npoints = 1;
	int errcode;
	
	FORTRAN_NAME(geovis)((double *) &latitude, (double *) &longitude, (int *) &npoints, azimut, zenithal, &errcode);

	if (errcode) return GEOSTAT_ERR_BAD_INIT;
	
	return GEOSTAT_ERR_OK;

}


