#!/usr/bin/perl

use strict;
use Getopt::Std;
use File::Basename;
use Time::Local;

my $options = 'Imvo:l:w:Fi:N:T:M:D:'; # with getopts

my $appname = $0;
my $version = '0.1';


if (! $ENV{COOKIE_CUTTER_ROOT}) {
	print "$0: \$COOKIE_CUTTER_ROOT: environment variable not defined\n";
	exit 1;
}
my $COOKIE_CUTTER_ROOT = $ENV{COOKIE_CUTTER_ROOT};

my %opt_flags;
my $node_lon = '+00';
my $node_incl = '+98.2';
my $slot_mean_time = '';
my $node_time = '';
my $width = '10';
my $output_dir = '.';
my $output_file;
my $msg_orbit = "$COOKIE_CUTTER_ROOT/bin/msg_orbit";
my $opt_multislot = 0;
my $ignore_time_infos = 0;

my %input_table = (
	'VIS006'           =>    'H-000-MSG1__-MSG1________-VIS006___-CYCLE____-',
	'VIS008'           =>    'H-000-MSG1__-MSG1________-VIS008___-CYCLE____-',
	'IR_016'           =>    'H-000-MSG1__-MSG1________-IR016____-CYCLE____-',
	'IR_039'           =>    'H-000-MSG1__-MSG1________-IR039____-CYCLE____-',
	'WV_062'           =>    'H-000-MSG1__-MSG1________-WV062____-CYCLE____-',
	'WV_073'           =>    'H-000-MSG1__-MSG1________-WV073____-CYCLE____-',
	'IR_087'           =>    'H-000-MSG1__-MSG1________-IR087____-CYCLE____-',
	'IR_097'           =>    'H-000-MSG1__-MSG1________-IR097____-CYCLE____-',
	'IR_108'           =>    'H-000-MSG1__-MSG1________-IR108____-CYCLE____-',
	'IR_120'           =>    'H-000-MSG1__-MSG1________-IR120____-CYCLE____-',
	'IR_134'           =>    'H-000-MSG1__-MSG1________-IR134____-CYCLE____-'
);

my %mono_time_table = (
	'+0000'            =>    +00
);

my %multi_time_table = (
	'-0060'            =>    -60,
	'-0045'            =>    -45,
	'-0030'            =>    -30,
	'-0015'            =>    -15,
	'+0000'            =>    +00,
	'+0015'            =>    +15,
	'+0030'            =>    +30,
	'+0045'            =>    +45,
	'+0060'            =>    +60
);

my $time_table = \%mono_time_table;

sub display_version {
    print STDERR "$appname: version $version\n";
} # display_version

sub parse_options {
	my $optargs = '';
    foreach (keys %opt_flags) {
		if ($_ eq 'l') {
			$node_lon = $opt_flags{$_};
		}
		elsif ($_ eq 'i') {
			$node_incl = $opt_flags{$_};
		}
		elsif ($_ eq 'o') {
			$output_file = $opt_flags{$_};
		}	
		elsif ($_ eq 'w') {
			$width = $opt_flags{$_};
		}
		elsif ($_ eq 'M') {
			$slot_mean_time = $opt_flags{$_};
		}
		elsif ($_ eq 'N') {
			$node_time = $opt_flags{$_};
		}
		elsif ($_ eq 'D') {
			$output_dir = $opt_flags{$_};
			next; # don't give this option to msg_orbit
		}
		elsif ($_ eq 'm') {
			$time_table = \%multi_time_table;
			$msg_orbit = "$COOKIE_CUTTER_ROOT/msg_orbit_multislot";
			$opt_multislot = 1;
			next; # don't give this option to msg_orbit
		}
		elsif ($_ eq 'F') {
			$optargs .= ' -F';
			next;
		}
		elsif ($_ eq 'I') {
			$optargs .= ' -I';
			$ignore_time_infos = 1;
			next;
		}
		$optargs .= " -$_ $opt_flags{$_}"; # these options (with mandatory arguments !) will be given to msg_orbit
    }
	return $optargs;
} # parse_options

sub echo_arguments {
    foreach (@ARGV) {
	print "$_\n";
    }
    exit 1;
} # echo_arguments

sub usage {
    print STDERR "usage: $appname [OPTIONS] input_directory slot_date\n";
	print STDERR "slot_date format: YYYYMMDDhhmm\n";
	exit 1;
}

sub parse_date($) {
	my $date = shift(@_);
	my $second = substr($date, -2, 2);
	my $minute = substr($date, -4, 2);
	my $hour   = substr($date, -6, 2);
	my $day    = substr($date, -8, 2);
	my $month  = substr($date, -10, 2);
	my $year   = substr($date, -14, 4);
	
	return ($second, $minute, $hour, $day, $month, $year);
}

sub date_to_time($) {
	my $date = shift(@_);
	my ($second, $minute, $hour, $day, $month, $year) = parse_date($date);
	my $time = timegm($second + 0, $minute + 0, $hour + 0, $day + 0, $month - 1, $year - 1900);
	return $time;
}

sub time_to_date($) {
	my $time = shift(@_);
	my ($second, $minute, $hour, $day, $month, $year) = gmtime($time);
	$month++;
	$year += 1900;
	$second = sprintf "%2.2d", $second;
	$minute = sprintf "%2.2d", $minute;
	$hour   = sprintf "%2.2d", $hour;
	$day    = sprintf "%2.2d", $day;
	$month  = sprintf "%2.2d", $month;
	$year   = sprintf "%2.2d", $year;
	my $date = "$year$month$day$hour$minute$second";
	
	return $date;
}

sub time_to_date_in_minutes($) {
	my $time = shift(@_);
	my ($second, $minute, $hour, $day, $month, $year) = gmtime($time);
	$month++;
	$year += 1900;
	$second = sprintf "%2.2d", $second; # unused
	$minute = sprintf "%2.2d", $minute;
	$hour   = sprintf "%2.2d", $hour;
	$day    = sprintf "%2.2d", $day;
	$month  = sprintf "%2.2d", $month;
	$year   = sprintf "%2.2d", $year;
	my $date = "$year$month$day$hour$minute";
	
	return $date;
}

# Main code

unless ( @ARGV > 0 ) {
    &usage;
    exit 1;
}

&getopts($options, \%opt_flags); # with Getopt::Std only
my $optargs = &parse_options;
#&echo_arguments;

usage() unless ($#ARGV == 1);

my $input_dir = $ARGV[0];
my $slot_date = $ARGV[1];

if ($slot_date !~ m/[0-9]{12}$/) { usage; }

my $slot_time = date_to_time($slot_date . '00');

if (length($slot_mean_time) == 0) {
	my $slot_mean_xtime = $slot_time + 450; # adds 7'30
	my $slot_xdate = time_to_date($slot_mean_xtime);
	$optargs .= " -t ${slot_xdate}";
}

if (length($node_time) == 0) {
	my $slot_mean_xtime = $slot_time + 450; # adds 7'30
	my $slot_xdate = time_to_date($slot_mean_xtime);
	$optargs .= " -N ${slot_xdate}";
}

if (length($output_file) == 0) {
	$output_file = "ICF-IVW_${slot_date}_lon=${node_lon}.hdf";
	$output_file = "$output_dir/$output_file";
	$optargs .= " -o $output_file";
}

unlink $output_file if ( -f $output_file);


while (my ($time_suffix, $time_offset) = each %$time_table) {
	my $time = $slot_time;
	
	if ($opt_multislot) {
		$time += 60*$time_offset; # $time_offset in minutes, $time in seconds
	}
	my $date = time_to_date_in_minutes($time);

	while (my ($dataset, $input_file) = each %input_table) {
		$input_file .= $date;
		$dataset .= "_$time_suffix" if ($opt_multislot);
		next unless (-f "$input_dir/$input_file"); # ignores missing files
		if ($ignore_time_infos and ($dataset eq 'VIS006')) {
			print STDERR "$0: time infos ignored in the SEVIRI file at the user demand (maybe these infos are missing or not fiable ?); SEVIRI time infos will be estimated by interpolation\n";
		}
		my $cmd = "$msg_orbit $optargs $input_dir/$input_file $dataset";
		print `$cmd`;
	}
}

exit 0;
