/* msg_extract.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <assert.h>
#include <time.h>

#include <hdf/mfhdf.h>

#include "version.h" /* APPNAME */
#include "geostat.h"
#include "msg_grid.h"
#include "msg_extract.h"
#include "opt_flags.h"
#include "sds_props.h"

#undef square
#define square(x) ( (x)*(x) )

#define HDFTYPE_MAXLEN 32

 /* Earth radius is non relevant here (we work only in lat, lon coordinates, with Earth supposed spherical)
  * another way to see it is to consider that we work in a unit system where Earth radius is 1
  *
  * note that you could put the real Earth radius value here. In fact, any non-zero value would be fine and
  * shouldn't change the results
  */
#define EARTH_RADIUS 1

#include "deg_rad_conv.h"

#define LATLON_UNDEF -999.

#define TIME_UNDEF -1
#define UINT8_UNDEF 0 // 255
#define UINT16_UNDEF 0 // 65535
#define UINT32_UNDEF -1

#define MSG_DIM_TRACK 0
#define MSG_DIM_XTRACK 1

#define WRITE_ERROR ( APPNAME ": failed to write data into %s (maybe previously defined with other dimensions or type)\n" )

extern long int lround(double x); /* should be included in math.h but is not, so we define the prototype ourselves ;-( */

/* converts a point(radius, lat, lon) into vector(x,y,z) */
static void bind_vector_to_geopoint(vector_t *vector, const geopoint_t *point) {

  assert(vector != NULL);

  vector->x = (point->radius)*cos(point->lat)*cos(point->lon);
  vector->y = (point->radius)*cos(point->lat)*sin(point->lon);
  vector->z = (point->radius)*sin(point->lat);
  vector->len = sqrt( vector->x*vector->x + vector->y*vector->y + vector->z*vector->z );

}

/* normalizes vector */
static void normalize_vector(vector_t *vector) {

  assert(vector != NULL);

  vector->x /= vector->len;
  vector->y /= vector->len;
  vector->z /= vector->len;
  vector->len = 1;

}

/* computes C as the cross_product (AxB) */
static void cross_product(const vector_t *A, const vector_t *B, vector_t *C) {

  assert(A != NULL);
  assert(B != NULL);

  assert(A->len != 0.);
  assert(B->len != 0.);

  C->x = A->y*B->z - A->z*B->y;
  C->y = A->z*B->x - A->x*B->z;
  C->z = A->x*B->y - A->y*B->x;
  C->len = sqrt( C->x*C->x + C->y*C->y + C->z*C->z );

}

/* returns a c-string (NUL-terminated array of chars) containing
 * a readable representation of an HDF type, given the data type code 
 * 
 * warning: this function is not reentrant since it uses a static char buffer (don't use it in multi-thread applications) 
 */
static char *hdf_cstr_type(const int32 data_type) {
  static char cstr_type[HDFTYPE_MAXLEN];

  switch(data_type) {
  case DFNT_CHAR8 : { strcpy(cstr_type, "DFNT_CHAR8") ; break ; }
  case DFNT_UCHAR8 : { strcpy(cstr_type, "DFNT_UCHAR8"); break ; }
  case DFNT_INT8 : { strcpy(cstr_type, "DFNT_INT8"); break ; }
  case DFNT_UINT8 : { strcpy(cstr_type, "DFNT_UINT8"); break ; }
  case DFNT_INT16 : { strcpy(cstr_type, "DFNT_INT16"); break ; }
  case DFNT_UINT16 : { strcpy(cstr_type, "DFNT_UINT16"); break ; }
  case DFNT_INT32 : { strcpy(cstr_type, "DFNT_INT32"); break ; }
  case DFNT_UINT32 : { strcpy(cstr_type, "DFNT_UINT32"); break ; }
  case DFNT_FLOAT32 : { strcpy(cstr_type, "DFNT_FLOAT32"); break ; }
  case DFNT_FLOAT64 : { strcpy(cstr_type, "DFNT_FLOAT64"); break ; }
  default : { strcpy(cstr_type, "UNKNOWN") ; break ; }
  }

  return cstr_type;

}

/* returns size of hdf types at execution */
static size_t hdf_sizeof(const int32 data_type) {
  
  switch(data_type) {
  case DFNT_CHAR8 : { return sizeof(char8) ; break ; }
  case DFNT_UCHAR8 : { return sizeof(uchar8); break ; }
  case DFNT_INT8 : { return sizeof(int8); break ; }
  case DFNT_UINT8 : { return sizeof(uint8); break ; }
  case DFNT_INT16 : { return sizeof(int16); break ; }
  case DFNT_UINT16 : { return sizeof(uint16); break ; }
  case DFNT_INT32 : { return sizeof(int32); break ; }
  case DFNT_UINT32 : { return sizeof(uint32); break ; }
  case DFNT_FLOAT32 : { return sizeof(float32); break ; }
  case DFNT_FLOAT64 : { return sizeof(float64); break ; }
  default : { fprintf(stderr,"%s: %d: unknown hdf type code: %ld\n", __FILE__, __LINE__, data_type); abort(); /* should never arrive there */ ; break ; }
  }
}

/* given M.lat on the track (and M.radius != 0), computes M.lon and (M.lin, M.col) in MSG grid
 */
static int msg_get_track_pixel(/* input */
			       const double incli /* in radians */, 
			       const double Tsat /* in seconds */, 
			       const double node_lon /* in radians */, 
			       const time_t node_time /* in seconds since the Unix Epoch (1-jan-1970) */, 
			       /* output */
			       geopoint_t *M) {

  geostat_err_t errcode;
  const double Tearth = 24*60*60; /* in seconds */

  assert(M != NULL);
  if (M->lat < -M_PI_2 || M->lat > M_PI_2) {
#ifndef NDEBUG
    fprintf(stderr, "%s: %d: unexpected value for M->lat: %f\n", __FILE__, __LINE__, M->lat*RAD2DEG);
    abort();
#endif
    return -1;
  }

  assert(M->radius != 0);
  
  if (fabs(sin(M->lat)) > fabs(sin(incli))) return -1;

  M->lon = atan2(sin(M->lat)/tan(incli), cos(asin(sin(M->lat)/sin(incli)))) 
    - Tsat/Tearth * asin(sin(M->lat)/sin(incli))
    + node_lon;

  assert(node_time > 0);
  if (incli >= 0) {
    M->time = node_time + ((signed short) lround( M->lat * Tsat / (2*M_PI) ));
  }
  else {
    M->time = node_time - ((signed short) lround( M->lat * Tsat / (2*M_PI) ));
  }
  errcode = geostat_latlon_to_lincol_double(M->lat, M->lon, &M->lin, &M->col);
  if (errcode == GEOSTAT_ERR_OUT_OF_RANGE || errcode == GEOSTAT_ERR_BAD_INPUT_COORD) return -1;
  assert(errcode == GEOSTAT_ERR_OK);

  assert(M->lon >= -M_PI && M->lon <= M_PI);
  
  return 0;

}

/* returns a track object (set of pixels on the track) given orbital parameters (in user_param) */
static int msg_get_track(const msg_user_param_t *user_param, track_t *track) {

  int ilat;
  int itrack;
  double ilat_per_deg = user_param->ilat_per_deg;
  double latmin = user_param->latmin;
  double latmax = user_param->latmax;
  int ilatmin = lround(latmin * ilat_per_deg);
  int ilatmax = lround(latmax * ilat_per_deg);
  const double incli = user_param->incli; /* in radians */
  const double Tsat  = user_param->Tsat; /* in seconds */
  const double node_lon = user_param->node_lon; /* in radians */
  const time_t node_time = user_param->node_time; /* in seconds since the Unix Epoch (1-jan-1970) */
  geopoint_t M;

  assert(user_param != NULL);
  assert(track != NULL);
  M.radius = EARTH_RADIUS;

  for (ilat = ilatmin, itrack = 0 ; ilat <= ilatmax ; ilat++) {
    M.lat = ilat/ilat_per_deg * DEG2RAD;
    if ((msg_get_track_pixel(incli, Tsat, node_lon, node_time, &M) < 0) || (itrack > 0 && lround(M.lin) == track->line[itrack - 1] )) continue;
    
    track->line[itrack] = lround(M.lin);
    track->column[itrack] = lround(M.col);
    track->lat[itrack] = M.lat;
    track->lon[itrack] = M.lon;
    track->time[itrack] = M.time;
    itrack++;
  }
  
  track->first = 0;
  track->last  = itrack - 1;

  if (track->last - track->first + 1 <= 0) return -1;
  return track->last - track->first + 1;

}

/* given a pixel column belonging to a crosstrack segment (between points W and E), returns its line */
static unsigned long crosstrack_line(const geopoint_t *E, const geopoint_t *W, const unsigned long column) {
  return lround(W->lin + (E->lin - W->lin)*(column - W->col)/(E->col - W->col));
}

/* returns a crosstrack object (set of pixels on the crosstrack), given user-defined width (in user_param),
 * intersection with the track (pixel M) and adjacent pixels on the track (previous pixel P and next pixel N)
 * vector PN (track-tangential) is needed to compute orthogonal direction to the track around M
 */
static int msg_get_crosstrack(const geopoint_t *M, const geopoint_t *N, const geopoint_t *P,
			      const msg_user_param_t *user_param, const void *data, crosstrack_t *crosstrack) {
  vector_t OM, OP, ON; /* current (M), previous (P) and next (N) pixel along the track (in x,y,z coordinates) */
  vector_t OE, OW; /* vectors for East and West crosstrack extremities */
  vector_t PN; /* PN = ON - OP */
  vector_t K; /* K = cross_product(PN, OM) */
  geopoint_t E, W; /* East and West crosstrack extremities */
  double sin_E_lat, sin_W_lat; /* computation variables (sinus of E.lat and W.lat) */
  const double half_width = user_param->half_width_in_radians;
  const int32 data_type = user_param->data_type;
  const int opt_flags = user_param->opt_flags;
  unsigned long (* get_line)(const geopoint_t *E, const geopoint_t *W, const unsigned long column);
  
  unsigned long line, column; /* line and column of the running pixel in the crosstrack */
  double lat, lon; /* latitude and longitude of the running pixel in the crosstrack */
  unsigned long first_column, last_column; /* columns of the first and last pixel (from west to east) in the crosstrack */
  unsigned long ixtrack; /* running index along the crosstrack */
  unsigned long nixtrack;
  double width_correction;
  geostat_err_t errcode;

  assert(user_param != NULL);
  assert(crosstrack != NULL);

  assert(M != NULL);
  assert(N != NULL);
  assert(P != NULL);

  bind_vector_to_geopoint(&OM, M);
  bind_vector_to_geopoint(&ON, N);
  bind_vector_to_geopoint(&OP, P);
 
  /* PN: tangent vector over the track, between P and N (around M) */
  
  PN.x = ON.x - OP.x;
  PN.y = ON.y - OP.y;
  PN.z = ON.z - OP.z;
  PN.len = sqrt( PN.x*PN.x + PN.y*PN.y + PN.z*PN.z );

  /* K: cross product between PN and OM (normal to the track and tangent to the Earth) */

  cross_product(&PN, &OM, &K);
  normalize_vector(&K);

  /* find crosstrack extremities E and W (half_width should be in radians) */

  if (opt_flags & OPT_WIDTH_CORRECTION) width_correction = cos(M->lat);
  else width_correction = 1.0;

  OE.x = OM.x + EARTH_RADIUS * half_width * width_correction * K.x;
  OE.y = OM.y + EARTH_RADIUS * half_width * width_correction * K.y;
  OE.z = OM.z + EARTH_RADIUS * half_width * width_correction * K.z;
  OE.len = sqrt( OE.x*OE.x + OE.y*OE.y + OE.z*OE.z );

  OW.x = OM.x - EARTH_RADIUS * half_width * width_correction * K.x;
  OW.y = OM.y - EARTH_RADIUS * half_width * width_correction * K.y;
  OW.z = OM.z - EARTH_RADIUS * half_width * width_correction * K.z;
  OW.len = sqrt( OW.x*OW.x + OW.y*OW.y + OW.z*OW.z );

  sin_E_lat = OE.z / OE.len;
  sin_W_lat = OW.z / OW.len;

  assert(sin_E_lat > -1 && sin_E_lat < 1);
  assert(sin_W_lat > -1 && sin_W_lat < 1);

  E.lat = asin(sin_E_lat);
  W.lat = asin(sin_W_lat);
  E.lon = atan2(OE.y, OE.x);
  W.lon = atan2(OW.y, OW.x);

  assert(E.lat > -M_PI_2 && E.lat < M_PI_2);
  assert(E.lon > -M_PI && E.lon < M_PI);

  errcode = geostat_latlon_to_lincol_double(E.lat, E.lon, &E.lin, &E.col);
  if (errcode == GEOSTAT_ERR_OUT_OF_RANGE) goto crosstrack_reset;
  
  assert(errcode == GEOSTAT_ERR_OK);

  assert(W.lat > -M_PI_2 && W.lat < M_PI_2);
  assert(W.lon > -M_PI && W.lon < M_PI);

  errcode = geostat_latlon_to_lincol_double(W.lat, W.lon, &W.lin, &W.col);
  if (errcode == GEOSTAT_ERR_OUT_OF_RANGE) goto crosstrack_reset;
  
  assert(errcode == GEOSTAT_ERR_OK);

  first_column = lround(W.col);
  last_column  = lround(E.col);
  get_line = &crosstrack_line;
 
  if (first_column < 0) first_column = 0;
  if (last_column >= MSG_NCOLUMNS) last_column = MSG_NCOLUMNS - 1;

  for (ixtrack = 0 ; ixtrack < CROSSTRACK_MAXSIZE ; ixtrack++) {
    int j;
    for (j = 0 ; j < CROSSTRACK_THICKNESS ; j++) {
      crosstrack->line[j][ixtrack] = UINT32_UNDEF;
      crosstrack->column[j][ixtrack] = UINT32_UNDEF;
      crosstrack->lat[j][ixtrack] = LATLON_UNDEF;
      crosstrack->lon[j][ixtrack] = LATLON_UNDEF;
      crosstrack->moving_sat_time[j][ixtrack] = TIME_UNDEF;
      crosstrack->seviri_acq_time[j][ixtrack] = TIME_UNDEF;
      crosstrack->data_uint8[j][ixtrack] = UINT8_UNDEF;
      crosstrack->data_uint16[j][ixtrack] = UINT16_UNDEF;
    }
  }

  nixtrack = last_column - first_column;
  if (nixtrack <= 0) nixtrack = 0;

  switch (data_type) {
  case (DFNT_UINT8) : {

    for (column = first_column, ixtrack = 0 ; column <= last_column ; column++, ixtrack++) {
      int j;
      for (j = 0 ; j < CROSSTRACK_THICKNESS ; j++) {
	line = get_line(&E, &W, column) + j - 1;
	if (line < 1 || line > MSG_NLINES) goto crosstrack_reset;
	errcode = geostat_lincol_to_latlon_double(line, column, &lat, &lon);
	if (errcode == GEOSTAT_ERR_OK) {
	  crosstrack->line[j][ixtrack] = line;
	  crosstrack->column[j][ixtrack] = column;
	  crosstrack->lat[j][ixtrack] = lat * RAD2DEG;
	  crosstrack->lon[j][ixtrack] = lon * RAD2DEG;
	  crosstrack->moving_sat_time[j][ixtrack] = M->time;
	  crosstrack->seviri_acq_time[j][ixtrack] = user_param->rows_infos->time[MSG_NLINES - line];
	  crosstrack->data_uint8[j][ixtrack] = ((uint8 *) data)[MSG_NCOLUMNS*(line - 1) + (column - 1)];
	  
	} else {
	  assert(errcode == GEOSTAT_ERR_OUT_OF_RANGE);
	  crosstrack->line[j][ixtrack] = UINT32_UNDEF;
	  crosstrack->column[j][ixtrack] = UINT32_UNDEF;
	  crosstrack->lat[j][ixtrack] = LATLON_UNDEF;
	  crosstrack->lon[j][ixtrack] = LATLON_UNDEF;
	  crosstrack->moving_sat_time[j][ixtrack] = TIME_UNDEF;
	  crosstrack->seviri_acq_time[j][ixtrack] = TIME_UNDEF;
	  crosstrack->data_uint8[j][ixtrack] = UINT8_UNDEF;
	}
      }
      
    } /* for (column) */
    break;
  } /* case (DFNT_UINT8) */
  case (DFNT_UINT16) : {

    for (column = first_column, ixtrack = 0 ; column <= last_column ; column++, ixtrack++) {
      int j;
      for (j = 0 ; j < CROSSTRACK_THICKNESS ; j++) {
	line = get_line(&E, &W, column) + j - 1;
	if (line < 1 || line > MSG_NLINES) goto crosstrack_reset;
	errcode = geostat_lincol_to_latlon_double(line, column, &lat, &lon);
	if (errcode == GEOSTAT_ERR_OK) {
	  crosstrack->line[j][ixtrack] = line;
	  crosstrack->column[j][ixtrack] = column;
	  crosstrack->lat[j][ixtrack] = lat * RAD2DEG;
	  crosstrack->lon[j][ixtrack] = lon * RAD2DEG;
	  crosstrack->moving_sat_time[j][ixtrack] = M->time;
	  crosstrack->seviri_acq_time[j][ixtrack] = user_param->rows_infos->time[MSG_NLINES - line];
	  crosstrack->data_uint16[j][ixtrack] = ((uint16 *) data)[MSG_NCOLUMNS*(line - 1) + (column - 1)];
	  
	} else {
	  assert(errcode == GEOSTAT_ERR_OUT_OF_RANGE);
	  crosstrack->line[j][ixtrack] = UINT32_UNDEF;
	  crosstrack->column[j][ixtrack] = UINT32_UNDEF;
	  crosstrack->lat[j][ixtrack] = LATLON_UNDEF;
	  crosstrack->lon[j][ixtrack] = LATLON_UNDEF;
	  crosstrack->moving_sat_time[j][ixtrack] = TIME_UNDEF;
	  crosstrack->seviri_acq_time[j][ixtrack] = TIME_UNDEF;
	  crosstrack->data_uint16[j][ixtrack] = UINT16_UNDEF;
	}
      }
    } /* for (column) */
    break;
  } /* case (DFNT_UINT16) */
  default : {
    fprintf(stderr, "%s: %d: fatal error: unexpected value for data_type: %s\n", __FILE__, __LINE__, hdf_cstr_type(data_type));
    abort(); /* should never arrive here in a non-buggy code */
    exit(EXIT_FAILURE);
  } /* default */
  } /* switch (data_type) */

  crosstrack->first = 0;
  crosstrack->last =  nixtrack;

  if (nixtrack <= 0) goto crosstrack_reset;
  
  return nixtrack; /* returns number of pixels in the crosstrack */

 crosstrack_reset:
  
  for (ixtrack = 0 ; ixtrack < CROSSTRACK_MAXSIZE ; ixtrack++) {
    int j;
    for (j = 0 ; j < CROSSTRACK_THICKNESS ; j++) {
      crosstrack->line[j][ixtrack] = UINT32_UNDEF;
      crosstrack->column[j][ixtrack] = UINT32_UNDEF;
      crosstrack->lat[j][ixtrack] = LATLON_UNDEF;
      crosstrack->lon[j][ixtrack] = LATLON_UNDEF;
      crosstrack->moving_sat_time[j][ixtrack] = TIME_UNDEF;
      crosstrack->seviri_acq_time[j][ixtrack] = TIME_UNDEF;
      crosstrack->data_uint8[j][ixtrack] = UINT8_UNDEF;
      crosstrack->data_uint16[j][ixtrack] = UINT16_UNDEF;
    }
  }
  return -1;

}

/* returns an extraction array (a filtered Seviri grid along a user-defined track)
 * given an input data array (Seviri grid) and user parameters (orbital parameters and extraction width) 
 */
int msg_extract_data(extract_t *extraction, const void *data, int number_of_sds, sds_props_t *sds, msg_user_param_t *user_param) {
  register int ilat; /* index of latitude (latitude-proportional integer, e.g. from -9000 to +9000) */
  crosstrack_t crosstrack; /* crosstrack object (set of pixels along the crosstrack) */

  intn status;
  int32 sd_id;
  int ntrack; /* number of pixels along the track */
  int nxtrack; /* number of pixels along the crosstrack */
  int itrack; /* running index along the track */
  double ilat_per_deg; /* resolution of ilat (number of ilat values in one degree of latitude) */
  track_t track; /* track object (set of pixels along the track) */
  geopoint_t M, N, P;
  int ret;
  int i, j;
  const int sds_index = user_param->sds_index;
  const char *data_file = user_param->data_file;
  const int32 data_type = user_param->data_type;
  xrit_rows_infos_t *rows_infos = user_param->rows_infos;
  const int opt_flags = user_param->opt_flags;
  const double incli = user_param->incli; /* in radians */
  const double node_lon = user_param->node_lon; /* in radians */
  const double Tsat = user_param->Tsat; /* in seconds */
  char cstr_data_type[HDFTYPE_MAXLEN]; 
  char cstr_sds_data_type[HDFTYPE_MAXLEN];
  long first_valid_row = -1;
  long last_valid_row = -1;
  time_t slot_first_time = -1;
  time_t slot_mean_time = -1;
  time_t slot_last_time = -1;
  time_t node_time = -1;
  int is_new_file;

  assert(rows_infos);

  if (opt_flags & OPT_NO_VALIDITY_CHECK) {
    /* if option -I is set, we ignore validity and time infos
     * this may be useful to process data where validity and time infos are missing (e.g. data in 2003)
     * in that case, all data will be supposed valid and provided a slot mean time is given,
     * time will be computed by linear interpolation over 12 minutes (estimated duration of a slot)
     */

    first_valid_row = 0;
    last_valid_row = XRIT_NROWS - 1;
    if (user_param->mean_time) {
      time_t first_time = user_param->mean_time - 450; /* e.g. if mean_time = 12h07min30s, first_time = 12h00min00s */
      for (i = first_valid_row ; i <= last_valid_row ; i++) {
	rows_infos->time[i] = first_time + 12*60*i/XRIT_NROWS; /* if first_time = 12h00min00s, time[i] will be interpolated between 12h00min00s and 12h12min00s */
      }
    }
  }
  else {
    for (i = 0 ; i < XRIT_NROWS ; i++) {
      if (rows_infos->validity[i] == 1) {
	first_valid_row = i;
	break;
      }
    }
    
    for (i = XRIT_NROWS - 1 ; i >= 0 ; i--) {
      if (rows_infos->validity[i] == 1) {
	last_valid_row = i;
	break;
      }
    }
  }

  if (first_valid_row == -1 || last_valid_row == -1) {
    fprintf(stderr, APPNAME ": no valid data found in %s\n", data_file);
    exit (EXIT_FAILURE);
  }
  
  if (user_param->mean_time) {
    slot_mean_time = user_param->mean_time;
  } 
  else {
    /* this is NOT a bug : slot_first_time does correspond to last_valid_row, and
     * slot_last_time to first_valid_row. Seviri sweeps Earth from south (bottom, last row, first time)
     * to north (top, first row, last time)
     */
    slot_first_time = rows_infos->time[last_valid_row];
    slot_last_time  = rows_infos->time[first_valid_row];
    slot_mean_time = (slot_first_time + slot_last_time) / 2;
    slot_mean_time = llround(slot_mean_time/450.)*450; // converts seconds into 7'30-periods (half a slot, 450 s)
  }
  
  if (slot_mean_time == -1) {
    fprintf(stderr, APPNAME ": no time info found in %s\n", data_file);
    exit (EXIT_FAILURE);
  }

  if (user_param->node_time) {
    node_time = user_param->node_time;
  }

  int32 compression_type = user_param->compression_type;
  int32 compression_level = user_param->compression_level;
  HDF_CHUNK_DEF chunk_def;

  strcpy(cstr_data_type, hdf_cstr_type(data_type));
  strcpy(cstr_sds_data_type, hdf_cstr_type(sds[sds_index].data_type));

  int32 *sds_id = malloc(number_of_sds*sizeof(int32));
  int32 start[2], edge[2];
  int32 dim_sizes[2];

  if (data_type != sds[sds_index].data_type) {
    fprintf(stderr, APPNAME ": type of data in %s (%s) incompatible with SDS %s (%s)\n", data_file, cstr_data_type,
	    sds[sds_index].name, cstr_sds_data_type);
    return -1;
  }
  
  ntrack = msg_get_track(user_param, &track);
  if (ntrack <= 0) {
    fprintf(stderr, "msg_extract: ntrack <= 0: maybe the longitude is not in [-90,+90]\n");
    return -1;
  }
  ilat_per_deg = user_param->ilat_per_deg;

  /* get equatorial crosstrack around node_lon = 0 */
  ilat = 0;

  M.lat = ilat/ilat_per_deg*DEG2RAD;
  M.radius = EARTH_RADIUS;
  
  assert(M.lat > -M_PI_2 && M.lat < M_PI_2);
  ret = msg_get_track_pixel(incli, Tsat, 0. /* node_lon */, node_time, &M);
  if (ret != 0) return -1;
  
  N.lat = (ilat+1)/ilat_per_deg*DEG2RAD;
  N.radius = EARTH_RADIUS;
    
  assert(N.lat > -M_PI_2 && N.lat < M_PI_2);
  ret = msg_get_track_pixel(incli, Tsat, 0. /* node_lon */, node_time, &N);
  if (ret != 0) return -1;
  
  P.lat = (ilat-1)/ilat_per_deg*DEG2RAD;
  P.radius = EARTH_RADIUS;
    
  assert(P.lat > -M_PI_2 && P.lat < M_PI_2);
  ret = msg_get_track_pixel(incli, Tsat, 0. /* node_lon */, node_time, &P);
  if (ret != 0) return -1;
  
  nxtrack = msg_get_crosstrack(&M, &N, &P, user_param, data, &crosstrack);
  if (nxtrack <= 0) return -1;

  if (ntrack > 0 && nxtrack > 0) {

    dim_sizes[MSG_DIM_TRACK] = ntrack * CROSSTRACK_THICKNESS;
    dim_sizes[MSG_DIM_XTRACK] = nxtrack;

    sd_id = SDstart(user_param->hdf_file, DFACC_WRITE); /* tries to update hdf_file */
    if (sd_id == FAIL) { /* if it fails, tries to create it ) */
      is_new_file = 1;

      sd_id = SDstart(user_param->hdf_file, DFACC_CREATE); if (status == FAIL) HEprint(stderr, 0); assert(sd_id != FAIL);

      for (i = 0 ; i < number_of_sds ; i++) {
	sds_id[i] = SDcreate(sd_id, sds[i].name, sds[i].data_type, sds[i].rank, dim_sizes); if (status == FAIL) HEprint(stderr, 0); assert(sds_id[i] != FAIL);

	if (compression_type == COMP_CODE_DEFLATE) {
	  assert(sds[i].rank == 2);
	  chunk_def.chunk_lengths[0] = chunk_def.comp.chunk_lengths[0] = dim_sizes[0];
	  chunk_def.chunk_lengths[1] = chunk_def.comp.chunk_lengths[1] = dim_sizes[1];
	  chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	  chunk_def.comp.cinfo.deflate.level = compression_level;

	  if (SDsetchunk(sds_id[i], chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	    fprintf(stderr, APPNAME ": cannot set chunks for SDS %s\n", sds[i].name);
	    exit(EXIT_FAILURE);
	  }

	} else if (compression_type == COMP_CODE_SKPHUFF) {
	  assert(sds[i].rank == 2);
	  chunk_def.chunk_lengths[0] = chunk_def.comp.chunk_lengths[0] = dim_sizes[0];
	  chunk_def.chunk_lengths[1] = chunk_def.comp.chunk_lengths[1] = dim_sizes[1];
	  chunk_def.comp.comp_type = COMP_CODE_DEFLATE;
	  chunk_def.comp.cinfo.skphuff.skp_size =  hdf_sizeof(sds[i].data_type);
	  
	  if (SDsetchunk(sds_id[i], chunk_def, HDF_CHUNK | HDF_COMP) == FAIL) {
	    fprintf(stderr, APPNAME ": cannot set chunks for SDS %s\n", sds[i].name);
	    exit(EXIT_FAILURE);
	  }
	} else {
	  if (compression_type != COMP_CODE_NONE) {
	    fprintf(stderr, APPNAME ": invalid or unknown compression type, compression ignored (only Adaptive Huffman and Lempel/Ziv-77 GZIP deflation are supported today)\n");
	    compression_type = COMP_CODE_NONE;
	  }
	} /* if/else (compression_type) */

      } /* for (i) */

    } else { /* updates hdf_file */
      is_new_file = 0;
      
      for (i = 0 ; i < number_of_sds ; i++) {
	sds_id[i] = SDselect(sd_id, i); assert(sds_id[i] != FAIL);
      }

    }
         
    status = SDsetcal(sds_id[sds_index], sds[sds_index].cal, sds[sds_index].cal_err, sds[sds_index].offset, sds[sds_index].offset_err, sds[sds_index].data_type);
    if (status == FAIL) HEprint(stderr, 0);

  } /* if (ntrack > 0 && nxtrack > 0) */

  start[MSG_DIM_XTRACK] = 0; /* writes data starting at the beginning of each cross track */
  edge[MSG_DIM_TRACK] = 1;  /* writes one crosstrack at a time */
  edge[MSG_DIM_XTRACK] = nxtrack; /* nxtrack pixels by crosstrack to write */

  for (itrack = track.first ; itrack <= track.last ; itrack++) {

    P.radius = EARTH_RADIUS;
    P.lat = track.lat[itrack - 1];
    P.lon = track.lon[itrack - 1];
    P.lin = track.line[itrack - 1];
    P.col = track.column[itrack - 1];

    M.radius = EARTH_RADIUS;
    M.lat = track.lat[itrack];
    M.lon = track.lon[itrack];
    M.lin = track.line[itrack];
    M.col = track.column[itrack];
    M.time = track.time[itrack];

    N.radius = EARTH_RADIUS;
    N.lat = track.lat[itrack + 1];
    N.lon = track.lon[itrack + 1];
    N.lin = track.line[itrack + 1];
    N.col = track.column[itrack + 1];

    msg_get_crosstrack(&M, &N, &P, user_param, data, &crosstrack);
 
    for (j = 0 ; j < CROSSTRACK_THICKNESS ; j++) {
      start[MSG_DIM_TRACK] = CROSSTRACK_THICKNESS*itrack + j; /* goes to the i-th crosstrack along the track */
      if (is_new_file) {
	status = SDwritedata(sds_id[0], start, NULL, edge, (VOIDP) crosstrack.moving_sat_time[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}
	
	status = SDwritedata(sds_id[1], start, NULL, edge, (VOIDP) crosstrack.seviri_acq_time[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}

	status = SDwritedata(sds_id[2], start, NULL, edge, (VOIDP) crosstrack.lat[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}
	status = SDwritedata(sds_id[3], start, NULL, edge, (VOIDP) crosstrack.lon[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}
	
	status = SDwritedata(sds_id[4], start, NULL, edge, (VOIDP) crosstrack.line[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}
	status = SDwritedata(sds_id[5], start, NULL, edge, (VOIDP) crosstrack.column[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}
      }
      
      switch (data_type) {
      case DFNT_UINT8 : {
	status = SDwritedata(sds_id[sds_index], start, NULL, edge, (VOIDP) crosstrack.data_uint8[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}
	break;
      }
      case DFNT_UINT16 : {
	status = SDwritedata(sds_id[sds_index], start, NULL, edge, (VOIDP) crosstrack.data_uint16[j]);
	if (status == FAIL) { 
	  fprintf(stderr, WRITE_ERROR, user_param->hdf_file); 
	  exit (EXIT_FAILURE);
	}
	break;
      }
      default : {
	fprintf(stderr, "%s: %d: fatal error: unexpected value for data_type: %s\n", __FILE__, __LINE__, hdf_cstr_type(data_type));
	abort(); /* should never arrive here in a non-buggy code */
	break;
      }
      } /* switch (data_type) */
      if (status == FAIL) HEprint(stderr, 0);
      assert(status != FAIL);
    } /* for j */  
  } /* for itrack */
    
  for (i = 0 ; i < number_of_sds ; i++) {
    status = SDendaccess(sds_id[i]); if (status == FAIL) HEprint(stderr, 0); assert(status != FAIL);
  }

  assert(sizeof(time_t) == sizeof(uint32));
  status = SDsetattr(sd_id, "slot_mean_time", DFNT_UINT32, 1, (uint32 *) &slot_mean_time); assert(status != FAIL);
  status = SDsetattr(sd_id, "orbit_node_time", DFNT_UINT32, 1, (uint32 *) &node_time); assert(status != FAIL);
  
  status = SDend(sd_id); assert(status != FAIL);
  
  return 0;

}
