/* msg_extract.h */

#ifndef _MSG_EXTRACT_H
#define _MSG_EXTRACT_H

#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <hdf/mfhdf.h> /* hdf types (uint8, uint16, uint32) */
#include "msg_grid.h"
#include "sds_props.h"
#include "xrit/include/libxrit.h"

#ifndef FILENAME_MAX /* is normally defined in stdio.h */
#define FILENAME_MAX 255 /**< Maximal length for filenames */
#endif

#define TRACK_MAXSIZE MSG_NLINES /**< maximum length of arrays defined in structure <i>_track_t</i> (maximum number of pixels along the track) */
#define CROSSTRACK_MAXSIZE MSG_NCOLUMNS /**< maximum length of arrays defined in structure <i>_crosstrack_t</i> (maximum number of pixels along one crosstrack) */
#define CROSSTRACK_THICKNESS 3 /**< to avoid loss of pixels during the extraction, crosstracks (which may be oblique) are given a thickness */

typedef float extract_t; /**< \deprecated the type of the extracted data */

/** \deprecated type for mode of extraction */
typedef enum _msg_extract_mode_t {
  MSG_EXTRACT_HORIZONTAL,
  MSG_EXTRACT_CROSSTRACK
} msg_extract_mode_t;

/** type for user parameters and data */
typedef struct _msg_user_param_t {
  double half_width_in_radians; /**< angular half width specified by the user (in radians) */
  double Tsat; /**< sun-synchronous orbit period, in minutes */
  double incli; /*< sun-synchronous orbit inclination, in radians */
  double node_lon; /**< sun-synchronous orbit node longitude, in radians (0:Greenwich, +:East, -:West) */
  time_t node_time; /**< sun-synchronous orbit node time (in seconds since 01-Jan-1970 00:00:00 TU) */
  time_t mean_time; /**< MSG/SEVIRI slot mean time (in seconds since 01-Jan-1970 00:00:00 TU) */
  unsigned long W_half_width; /**< equatorial crosstrack western half width (in pixels in the MSG/SEVIRI grid), the two crosstracks halves may not have the same widths in the MSG/SEVIRI grid */
  unsigned long E_half_width; /**< equatorial crosstrack eastern half width (in pixels in the MSG/SEVIRI grid), the two crosstracks halves may not have the same widths in the MSG/SEVIRI grid */
  int opt_flags; /**< options (for example, may be used by functions to know if verbose mode is enabled) */
  double latmin; /**< minimal (southest) latitude of extraction, in degrees */
  double latmax; /**< maximal (northest) latitude of extraction, in degrees */
  int ilat_per_deg; /**< number of latitude strides per degree (along the track) */
  int sds_index; /**< index of dataset (for loops over datasets) */
  char hdf_file[FILENAME_MAX]; /**< name of the output hdf file */
  char *data_file; /**< name of the input data file (only XRIT and FIS file formats are supported) */
  int32 data_type; /**< type of the input data (uint16 for XRIT files, uint8 or uint16 for FIS files) */
  int32 compression_type; /**< compression type: COMP_CODE_NONE (no compression), COMP_CODE_DEFLATE (gzip), COMP_CODE_SKPHUFF (huffman) */
  int32 compression_level; /**< compression level: 0-9, gzip only */
  msg_extract_mode_t extract_mode; /**< \deprecated extraction mode: MSG_EXTRACT_HORIZONTAL or MSG_EXTRACT_CROSSTRACK */
  xrit_rows_infos_t *rows_infos; /**< MSG/SEVIRI rows validity and time infos (see xrits_rows_infos_t in xrit/include/read_xrit.h) */
} msg_user_param_t;

/** representation of a pixel in MSG/SEVIRI and lat-lon 2D-grids */
typedef struct _pixel_t {
  unsigned long line; /**< line in the MSG/SEVIRI grid */
  unsigned long column; /**< column in the MSG/SEVIRI grid */
  double lat; /**< latitude */
  double lon; /**< longitude */
} pixel_t;

/** representation of a vector in a cartesian system coordinates */
typedef struct _vector_t {
  double x; /**< x-coordinate */
  double y; /**< y-coordinate */
  double z; /**< z-coordinate */
  double len; /**< length of the vector */
} vector_t;

/** representation of a geographic point in 3D-spherical coordinates + time coordinate */
typedef struct _geopoint_t {
  double radius; /**< Earth radius (in fact it could be any non-null value, but should be the same for all the points) */
  double lat; /**< latitude */
  double lon; /**< longitude */
  double lin; /**< line in the MSG/SEVIRI grid */
  double col; /**< column in the MSG/SEVIRI grid */
  time_t time; /**< expected view time by the sun-synchronous satellite */
} geopoint_t;

/** representation of a sun-synchronous satellite track */
typedef struct _track_t {
  unsigned long first; /**< first valid index of the following arrays */
  unsigned long last;  /**< last valid index of the following arrays */
  uint16 line[TRACK_MAXSIZE]; /**< lines of pixels belonging to the track (only those between <i>first</i> and <i>last</i> indexes are valid) */
  uint16 column[TRACK_MAXSIZE]; /**< columns of pixels belonging to the track (only those between <i>first</i> and <i>last</i> indexes are valid) */
  double lat[TRACK_MAXSIZE];  /**< latitudes of pixels belonging to the track (only those between <i>first</i> and <i>last</i> indexes are valid) */
  double lon[TRACK_MAXSIZE];  /**< longitudes of pixels belonging to the track (only those between <i>first</i> and <i>last</i> indexes are valid) */
  time_t time[TRACK_MAXSIZE];  /**< expected view time of pixels belonging to the track (only those between <i>first</i> and <i>last</i> indexes are valid) */
} track_t;

/** representation of a sun-synchronous satellite crosstrack */
typedef struct _crosstrack_t {
  unsigned long first; /**< first valid index of the following arrays (2nd dimension only) */
  unsigned long last; /**< last valid index of the following arrays (2nd dimension only) */
  uint16 line[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE]; /**< lines of pixels belonging to the crosstrack (only those between <i>first</i> and <i>last</i> indexes are valid) */
  uint16 column[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE]; /**< columns of pixels belonging to the crosstrack (only those between <i>first</i> and <i>last</i> indexes are valid) */
  float lat[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE];  /**< latitudes of pixels belonging to the crosstrack (only those between <i>first</i> and <i>last</i> indexes are valid) */
  float lon[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE];  /**< longitudes of pixels belonging to the crosstrack (only those between <i>first</i> and <i>last</i> indexes are valid) */
  time_t moving_sat_time[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE];  /**< expected time of view of crosstrack pixels (only those between <i>first</i> and <i>last</i> indexes are valid) 
								      * by the moving satellite */
  time_t seviri_acq_time[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE]; /**< time of acquisition of crosstrack pixels by SEVIRI (only those between <i>first</i> and <i>last</i> indexes are valid) */
  uint8 data_uint8[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE]; /**< uint8 data (seviri counts) of pixels belonging to the crosstrack (only those between <i>first</i> and <i>last</i> indexes are valid) */
  uint16 data_uint16[CROSSTRACK_THICKNESS][CROSSTRACK_MAXSIZE]; /**< uint16 data (seviri counts) of pixels belonging to the crosstrack (only those between <i>first</i> and <i>last</i> indexes are valid) */
} crosstrack_t;

/**
 * @param extraction an array containing the result of extraction. Must be allocated by the caller (with <b>MSG_NLINES*MSG_NCOLUMNS*sizeof(extract_t)</b> bytes)
 * @param data the input data to extract (an array of MSG_LINES*MSG_COLUMNS elements, whose type must be given by the field <b>data_type</b> of <i>param</i>)
 * @param number_of_sds size of the <b>sds</b> array
 * @param sds array of the datasets to write into the file <b>hdf_file</b> of <i>param</i>
 * @param param a structure that wraps user-defined parameters and other infos (to facilitate parameter passing)
 */
extern int msg_extract_data(extract_t *extraction, const void *data, int number_of_sds, sds_props_t *sds, msg_user_param_t *param);


#endif
