/* mosaic_seviri.cpp */

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <cassert>
#include <cfloat>
#include <cmath>
#include <new>
#include "Hdf.hpp"

#define UINT16_MAX 65535

#define MOSAIC_NROWS 3712
#define MOSAIC_NCOLS 3712

#define UNDEF_SLOPE 0.
#define UNDEF_OFFSET 0.

using namespace std;

int32 mosaic_start[] = { 0, 0 };
int32 mosaic_dims[] = { MOSAIC_NROWS, MOSAIC_NCOLS };
int opt_flags;

#ifndef FILENAME_MAX /* should be defined in cstdio */
#define FILENAME_MAX 255
#endif

#define OPT_VERBOSE                   0x00000001
#define OPT_DISABLE_CALIBRATION       0x00000002
#define DEFAULT_OUTPUT_FILENAME "seviri_mosaic.hdf"
char output_filename[FILENAME_MAX];
char raw_prefix[FILENAME_MAX];

int32 compression_type = COMP_CODE_NONE;
int32 compression_level = DEFAULT_COMP_LEVEL;

Hdf_file::sds_infos_type sds_infos[] = {
  { "Moving_Sat_Time",  DFNT_UINT32,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "Seviri_Acq_Time",  DFNT_UINT32,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "Latitude",         DFNT_FLOAT32, 2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "Longitude",        DFNT_FLOAT32, 2, mosaic_dims, 1., 0., 0., 0., 0, NULL },

  { "VIS006",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "VIS008",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "IR_016",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "IR_039",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "WV_062",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "WV_073",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "IR_087",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "IR_097",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "IR_108",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "IR_120",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL },
  { "IR_134",           DFNT_UINT16,  2, mosaic_dims, 1., 0., 0., 0., 0, NULL }
};

float64 *mosaic_slope;
float64 *mosaic_offset;

static const size_t nsds = sizeof(sds_infos)/sizeof(*sds_infos);

static Hdf_file *mosaic_file = NULL;

/* each mosaic tile (input files) comes with its individual calibration factors (for each channel)
 * in order to build the mosaic, we have to choose common calibration factors to keep with the mosaic
 * we keep the greatest slope for each channel, and the corresponding offset
 * Results are stored in the global array mosaic_slope and mosaic_offset
 *
 * This common factors (called mosaic calibration factors) will be used in process_dataset
 *
 * called by init_mosaic_calibration
 */

void compute_common_calibration(int nfiles, char **filenames) {

  for (unsigned int isds = 0 ; isds < nsds ; isds++) {
    char *dataset = sds_infos[isds].sds_name;
  
    for (int ifile = 0 ; ifile < nfiles ; ifile++) {
      float64 cal_slope;
      float64 cal_offset;
      
      Hdf_file *input_file = new Hdf_file(filenames[ifile], DFACC_READ);
      Hdf_sds sds = input_file->get_sds(dataset);
      try {
	sds.get_attribute("scale_factor").get_value(&cal_slope);
	if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: compute_common_calibration: " << filenames[ifile] << ": " << dataset << ": cal_slope = " << cal_slope << endl;
	sds.get_attribute("add_offset").get_value(&cal_offset);
	if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: compute_common_calibration: " << filenames[ifile] << ": " << dataset << ": cal_offset = " << cal_offset << endl;
	assert(cal_slope > 0.);
	if (mosaic_slope[isds] < cal_slope) {
	  mosaic_slope[isds] = cal_slope;
	  mosaic_offset[isds] = cal_offset;
	}
      }
      catch(string err_msg) { /* 'scale_factor' or 'add_offset' missing (e.g. Time, Latitude, Longitude) */
	if (strcmp(sds_infos[isds].sds_name, "Latitude") == 0
	    || strcmp(sds_infos[isds].sds_name, "Longitude") == 0
	    || strcmp(sds_infos[isds].sds_name, "Moving_Sat_Time") == 0
	    || strcmp(sds_infos[isds].sds_name, "Seviri_Acq_Time") == 0) goto next_sds;
	else cerr << err_msg << endl;
      }
      
      delete input_file;
    } // for (ifile)
    
    if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: compute_common_calibration: " << dataset 
				      << ": cal_slope = " << mosaic_slope[isds]
				      << " cal_offset = " << mosaic_offset[isds]
				      << endl; 

  next_sds:
    ;
  } // for (isds)
}

/* some initialization stuff
 *
 * called by main
 */

void init_mosaic_calibration(int nfiles, char **filenames) {

  mosaic_slope = new(nothrow) float64[nsds]; assert(mosaic_slope);
  mosaic_offset = new(nothrow) float64[nsds]; assert(mosaic_offset);

  for (unsigned int isds = 0 ; isds < nsds ; isds++) {
    mosaic_slope[isds] = UNDEF_SLOPE;
    mosaic_offset[isds] = UNDEF_OFFSET;
  }
  compute_common_calibration(nfiles, filenames);

}

/*
 * recalibrates each input data with the mosaic calibration factors (mosaic_slope and mosaic_offset) determined by compute_common_calibration
 *
 * called by build_mosaic (with each input file)
 */

template<typename datatype>
void process_dataset(Hdf_file *tile_file, const int isds, int32 tile_nrows, int32 tile_ncols, uint16 *row_buffer, uint16 *col_buffer, datatype *mosaic_buffer) {

  int32 tile_start[2] = { 0, 0 };
  int32 tile_edge[2];
  datatype *tile_buffer;
  float64 cal_slope = 1.0;
  float64 cal_offset = 0.0;
  char *dataset = sds_infos[isds].sds_name;
      
  assert(tile_file);
  assert(mosaic_file);
  assert(mosaic_buffer);

  tile_edge[0] = tile_nrows;
  tile_edge[1] = tile_ncols;
  tile_buffer = new datatype[tile_nrows*tile_ncols];
  Hdf_sds sds = tile_file->get_sds(dataset);
  sds.read(tile_buffer, tile_start, NULL, tile_edge);

  try {
    sds.get_attribute("scale_factor").get_value(&cal_slope);
    if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: process_dataset: " << dataset << " cal_slope = " << cal_slope << endl;
    sds.get_attribute("add_offset").get_value(&cal_offset);
    if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: process_dataset: " << dataset << " cal_offset = " << cal_offset << endl;
  }
  catch (string err_msg) { /* 'scale_factor' or 'add_offset' missing (e.g. Time, Latitude, Longitude) */
    if (strcmp(dataset, "Latitude") == 0
	|| strcmp(dataset, "Longitude") == 0
	|| strcmp(dataset, "Moving_Sat_Time") == 0
	|| strcmp(dataset, "Seviri_Acq_Time") == 0) /* do nothing */ ;
    else cerr << err_msg << endl;
  }

  for (int tile_irow = 0 ; tile_irow < tile_nrows ; tile_irow++) {
    for (int tile_icol = 0 ; tile_icol < tile_ncols ; tile_icol++) {
      uint16 seviri_irow;
      uint16 seviri_icol;

      seviri_irow = row_buffer[tile_irow*tile_ncols + tile_icol];
      if (seviri_irow == UINT16_MAX) continue;
      seviri_icol = col_buffer[tile_irow*tile_ncols + tile_icol];
      if (seviri_icol == UINT16_MAX) continue;

      seviri_irow--;
      seviri_icol--;

      if (mosaic_slope[isds] != UNDEF_SLOPE || (opt_flags & OPT_DISABLE_CALIBRATION == 0)) {
	float64 tmp;
	tmp = cal_slope*tile_buffer[tile_irow*tile_ncols + tile_icol] + cal_offset;
	mosaic_buffer[seviri_irow*MOSAIC_NCOLS + seviri_icol] = static_cast<datatype>(lround((tmp - mosaic_offset[isds])/mosaic_slope[isds]));
      } 
      else {
	mosaic_buffer[seviri_irow*MOSAIC_NCOLS + seviri_icol] = tile_buffer[tile_irow*tile_ncols + tile_icol];
      }
    }
  }
  
  delete tile_buffer;

}

/* builds one dataset (of index 'isds') of type <datatype> in the mosaic file
 * the input data are read from 'filenames'
 *
 * called by main for each dataset
 */

template<typename datatype>
void build_mosaic(int nfiles, char **filenames, const int isds) {

  int32 tile_start[2] = { 0, 0 };
  int32 tile_edge[2];
  int32 tile_nrows, tile_ncols;
  uint16 *row_buffer;
  uint16 *col_buffer;
  datatype *mosaic_buffer;
  char *dataset = sds_infos[isds].sds_name;

  mosaic_buffer = (datatype *) calloc(MOSAIC_NROWS*MOSAIC_NCOLS, sizeof(datatype));
  if (mosaic_buffer == NULL) {
    perror("mosaic_buffer = calloc()");
  }

  for (int i = 1 ; i <= nfiles ; i++) {
    vector<Hdf_file::sds_infos_type> sds_infos;
    Hdf_file *tile_file = new Hdf_file(filenames[i - 1], DFACC_READ, sds_infos);
    Hdf_sds sds_time  = tile_file->get_sds("Moving_Sat_Time");
    sds_time.get_dimensions(tile_edge);
    tile_nrows = tile_edge[0];
    tile_ncols = tile_edge[1];
    
    row_buffer = new(nothrow) uint16[tile_nrows*tile_ncols]; assert(row_buffer);
    col_buffer = new(nothrow) uint16[tile_nrows*tile_ncols]; assert(col_buffer);
     
    tile_file->read_sds("Row", row_buffer, tile_start, NULL, tile_edge);
    tile_file->read_sds("Column", col_buffer, tile_start, NULL, tile_edge);
    
    process_dataset(tile_file, isds, tile_nrows, tile_ncols, row_buffer, col_buffer, mosaic_buffer);
    delete tile_file;
    delete[] col_buffer;
    delete[] row_buffer;
    
  }

  if (mosaic_slope[isds] != UNDEF_SLOPE) {
    float64 hdf_slope = mosaic_slope[isds];
    float64 hdf_offset = mosaic_offset[isds];
    if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: " << dataset << ": hdf_slope = " << hdf_slope << " hdf_offset = " << hdf_offset << endl;
    mosaic_file->write_sds(dataset, mosaic_buffer, mosaic_start, NULL, mosaic_dims, hdf_slope, hdf_offset);
  }
  else {
    if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: " << dataset << " not calibrable" << endl;
    mosaic_file->write_sds(dataset, mosaic_buffer, mosaic_start, NULL, mosaic_dims);
  }
  while (*raw_prefix) {
    char raw_file[FILENAME_MAX];
    FILE *raw_stream;

    snprintf(raw_file, FILENAME_MAX, "%s_%s.raw", raw_prefix, sds_infos[isds].sds_name);
    if ((raw_stream = fopen(raw_file, "w")) == NULL) {
      perror(raw_file);
      break;
    }
    if (fwrite(mosaic_buffer, MOSAIC_NROWS*MOSAIC_NCOLS*sizeof(datatype), 1, raw_stream) != 1) {
      perror(raw_file);
      break;
    }
    if (fclose(raw_stream) != 0) {
      perror(raw_file);
      break;
    }
    break; /* yes, there is really a 'break' here (we use 'while' like an 'if' in order to use 'break' to leave the block when needed)*/
  }

  free(mosaic_buffer);

}

void usage(char *argv[]) {

  cerr << "usage: " << argv[0] << " [OPTIONS] tile1 [tile2...]" << endl;
  cerr << endl;
  cerr << "OPTIONS: " << endl;
  cerr << "  -o <output_file>        sets the output filename" << endl;
  cerr << "  -v                      verbose mode" << endl;
  cerr << "  -D                      disables calibration (for debugging purpose)" << endl;
  cerr << "  -c <compression type>   compression type may be : none, gzip or huffman (case insensitive). Default is none." << endl;
  cerr << "  -z <compression level>  compression level, relevant only for gzip: 0-9 (default is " << DEFAULT_COMP_LEVEL << ")\n" << endl;
	   
  exit (EXIT_FAILURE);

}

void parse_options(int *argc, char **argv[]) {
  int option;
  const char *options_list = "c:r:o:z:Dv";

  opt_flags = 0;
  
  while ((option = getopt(*argc, *argv, options_list)) != -1 ) {
    switch(option) {
    case 'D' :
      opt_flags |= OPT_DISABLE_CALIBRATION;
      break;
    case 'c' :
      if (strcasecmp(optarg, "none") == 0) compression_type = COMP_CODE_NONE;
      else if (strcasecmp(optarg, "gzip") == 0) compression_type = COMP_CODE_DEFLATE;
      else if (strcasecmp(optarg, "huffman") == 0) compression_type = COMP_CODE_SKPHUFF;
      else {
	fprintf(stderr, "%s: invalid value for compression type (option -c): %s (should be 'none', 'gzip' or 'huffman')\n", argv[0], optarg);
	exit (EXIT_FAILURE);
      }
      break;
    case 'r' :
       /* specifies raw prefix */
      strncpy(raw_prefix, optarg, FILENAME_MAX - 1);
      break;
    case 'o' :
      /* specifies output filename */
      strncpy(output_filename, optarg, FILENAME_MAX - 1);
      break;
    case 'v' :
      /* enables verbose mode */
      opt_flags |= OPT_VERBOSE;
      break;
    case 'z' :
      compression_level = atoi(optarg);
      if (compression_level < 0 || compression_level > 9) {
	fprintf(stderr, "%s: invalid value for compression level: %ld (should be between 0 and 9)\n", argv[0], compression_level);
	exit (EXIT_FAILURE);
      }
      break;
    case '?' :
      fprintf(stderr,"option %c : unknown\n\n", optopt);
      usage(argv[0]);
    default :
      fprintf(stderr, "%s: parse options : fatal error\n", argv[0]);
      exit(8);
    } /* switch */
  } /* while */
  *argc -= optind - 1;
  *argv += optind - 1;
}


int main(int argc, char *argv[]) {

  int nfiles = 0;
  char **filenames;

  parse_options(&argc, &argv);
  if (*output_filename == '\0') strcpy(output_filename, DEFAULT_OUTPUT_FILENAME);
  if (argc < 2) usage(argv);
  nfiles = argc - 1;

  filenames = new (char *)[nfiles];
  for (int i = 1 ; i <= nfiles ; i++) filenames[i - 1] = argv[i];

  init_mosaic_calibration(nfiles, filenames);
  mosaic_file = new Hdf_file(output_filename, DFACC_CREATE, sds_infos, nsds, SD_FILL, NULL, compression_type, compression_level);
  assert(mosaic_file);

  if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: Moving_Sat_Time" << endl;
  build_mosaic<uint32>(nfiles, filenames, 0 /* Moving_Sat_Time */);
  if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: Seviri_Acq_Time" << endl;
  build_mosaic<uint32>(nfiles, filenames, 1 /* Data_View_Time */);
  if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: Latitude" << endl;
  build_mosaic<uint32>(nfiles, filenames, 2 /* Latitude */);
  if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: Longitude" << endl;
  build_mosaic<uint32>(nfiles, filenames, 3 /* Longitude */);

  for (size_t isds = 4 ; isds < nsds ; isds++) {
    if (opt_flags & OPT_VERBOSE) cout << "msg_mosaic: " << sds_infos[isds].sds_name << endl;
    switch (sds_infos[isds].sds_type) {
    case DFNT_UINT8 : 
      build_mosaic<uint8>(nfiles, filenames, isds);
      break;
    case DFNT_UINT16 :
      build_mosaic<uint16>(nfiles, filenames, isds);
      break;
    default :
      cerr << "msg_mosaic: " << sds_infos[isds].sds_name << ": unexpected value for sds_type: " << sds_infos[isds].sds_type << endl;
      break;
    }
  }
  
  delete[] filenames;
  delete[] mosaic_slope;
  delete[] mosaic_offset;
  delete mosaic_file;

  return EXIT_SUCCESS;
}
