/* msg_orbit.c */

#define _XOPEN_SOURCE /**< for strptime (GlibC2 time.h) */
#include <time.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <assert.h>
#include <hdf/mfhdf.h> /* hdf types and compression modes */
#include "version.h"
#include "geostat.h"
#include "read_any.h"
#include "msg_extract.h"
#include "deg_rad_conv.h"
#include "opt_flags.h"
#include "msg_sds.h"
#include "custom_timegm.h"

extern int strcasecmp(const char *, const char *); /**< insensitive-case comparison between two strings (BSD extension) */

#ifndef FILENAME_MAX /* is normally defined in stdio.h */
#define FILENAME_MAX 255 /**< Maximum length of filenames */
#endif

#define PATHNAME_SEP '/' /**< path separator ('/' on Unix-like systems, '\' on DOS-like and Windows systems) */

#define DEFAULT_NODE_LON_IN_DEGREES 0. /**< Default longitude of the heliosynchronous orbit node, in degrees */
#define DEFAULT_INCLI_IN_DEGREES 98.2 /**< Default inclination of the heliosynchronous orbit, in degrees (90 = polar orbit) */
#define DEFAULT_TSAT_IN_MIN 98.88 /**< Default period of the heliosynchronous orbit, in minutes (should be negative for descending orbits from north to south) */
#define DEFAULT_LATMIN -85 /**< Default minimal latitude for extraction */
#define DEFAULT_LATMAX 85 /**< Default maximal latitude for extraction */
#define DEFAULT_WIDTH 10. /**< Default width of extraction (in degrees) */
#define DEFAULT_COMPRESSION_LEVEL 4 /**< Default compression level (for gzip only) */
#define DEFAULT_EXTRACT_MODE MSG_EXTRACT_CROSSTRACK /**< \deprecated Default extraction mode */
#define DEFAULT_ILAT_PER_DEG 20 /**<
				 * Default number of latitude strides (ilat) per degree.
				 * This should be near from the optimal value
				 * (lesser values should cause a loss of pixels, greater values
				 * should cause a time overload without gain of information)
				 * the optimal value depends on CROSSTRACK_THICKNESS defined in msg_extract.h
				 */

static char hdf_pathname[FILENAME_MAX]; /**< Name of the output HDF filename */
static char *hdf_file_extension = ".hdf"; /**< Extension of the output file */
static double half_width_in_degrees = DEFAULT_WIDTH / 2.; /**< Half-width of extraction (in degrees) */
static double node_lon_in_degrees = DEFAULT_NODE_LON_IN_DEGREES; /**< Longitude of the heliosynchronous orbit node */
static double incli = DEFAULT_INCLI_IN_DEGREES*DEG2RAD; /**< Orbital inclination of the heliosynchronous satellite */
static double Tsat = DEFAULT_TSAT_IN_MIN*60; /**< Orbital period, in seconds */
static double latmin = DEFAULT_LATMIN; /**< Minimal latitude of extraction */
static double latmax = DEFAULT_LATMAX; /**< Maximum latitude of extraction */
static int ilat_per_deg = DEFAULT_ILAT_PER_DEG; /**< Number of latitude strides (ilat) per degree, in the extraction process */
static int32 compression_type = COMP_CODE_NONE; /**< Type of compression, may be COMP_CODE_NONE, COMP_CODE_DEFLATE (gzip) or COMP_CODE_SKPHUFF (huffman) */
static int32 compression_level = DEFAULT_COMPRESSION_LEVEL; /**< Compression level (for gzip only) */
static int sds_index = -1; /**< Index of dataset (for loops over datasets) */
static msg_extract_mode_t extract_mode = DEFAULT_EXTRACT_MODE; /**< \deprecated Extraction mode */
static time_t node_time = -1; /**< Time of the heliosynchronous orbit node (in seconds since 01/01/1970,00:00 UTC) */
static time_t mean_time = -1; /**< Mean time of the SEVIRI slot to process (in seconds since 01/01/1970,00:00 UTC) */

static const char time_format[] = "%Y%m%d%H%M%S"; /**< expected format for read user-given times (with options -N and -M) */
static void *counts; /**< seviri radiances counts (pointer to a buffer allocated by the <i>read_any</i> function)
		      * actual type of counts should be:
		      * - uint16 for XRIT input files
		      * - uint8 or uint16 or FIS input files
		      * All of this is handled by <i>read_any</i>
		      * 
		      * counts is output from <i>read_any</i>, and input for <i>msg_extract_data</i>
		      */
static extract_t extraction[MSG_NLINES*MSG_NCOLUMNS]; /**< \deprecated
						       *
						       * an array in the MSG/SEVIRI projection containing the result of the extraction
						       * by msg_extract_data. Was stored in a raw data file in the early times of the development
						       * (for debugging purposes), not used anymore. To be removed.
						       * output from <i>msg_extract_data</i>
						       */

extern time_t timegm (struct tm *); /**< converts a struct tm into time_t (reverse of gmtime)
				     * FIXME: GNU extension (a more portable way to compute UT time conversion is described in 'timegm' man page)
				     */

/** @brief command usage, on-line help called when the command is misused */
void usage() {
  fprintf(stderr,"usage: " APPNAME " [OPTIONS] <MSG input file> <SDS name>\n\n");
  fprintf(stderr,"OPTIONS:\n");
  fprintf(stderr,
	  "  -o <hdf_file>             name of the output hdf file (add .hdf to the input file by default)\n"
	  "                            (uses the same name as the hdf default file, with .raw instead of .hdf, or add .raw if -o is used)\n"
	  "  -v                        verbose mode\n"
	  /* undocumented options
	  "  -L <latmin>,<latmax>      latitude range of the extraction (in degrees), default is %+d,%+d (not yet implemented)\n"
	  "  -W                        applies cos(latitude) as an extraction width correction (to have the extraction nearly constant in degrees)\n"
	  */
	  "  -I                        ignores validity check on data files\n"
	  "  -l <node longitude>       node longitude (in degrees, default is %g)\n"
	  "  -T <orbital period>       orbital period (in minutes, default is %g)\n"
	  "  -i <orbital inclination>  orbital inclination (in degrees, default is %g)\n"
	  "  -N <node time>            specifies the orbital node time\n"
	  "  -M <mean time>            specifies the Seviri slot mean time (may be useful when this info is missing or is not trustful in the Seviri file)\n"
	  "  -x <extraction mode>      sets the extraction mode: c)rosstrack (default), h)orizontal\n"
	  "  -w <width>                width of extraction around the track, in degrees (default is %g)\n"
	  /* undocumented option
	  "  -f <extraction factor>    the extraction factor is a positive integer (1 or more, default is %d)\n"
	  "                            controls the resolution in latitude along the track, time computation and size of the output hdf file\n"
	  "                            (for instance, 1 corresponds to 1 in latitude along the track and 100 to 1/100)\n"
	  */
	  "  -c <compression type>     compression type may be : none, gzip or huffman (case insensitive). Default is none.\n"
	  "  -z <compression level>    compression level, relevant only for gzip: 0-9 (default is %d)\n"
	  "\n",
	  /* DEFAULT_LATMIN, DEFAULT_LATMAX, */ DEFAULT_NODE_LON_IN_DEGREES, DEFAULT_TSAT_IN_MIN, DEFAULT_INCLI_IN_DEGREES, DEFAULT_WIDTH,
	  DEFAULT_COMPRESSION_LEVEL /* , DEFAULT_ILAT_PER_DEG */
	  );
  return;
}

/** @brief returns the index of a dataset whose name is given
 *
 * @param sds_name the name of a dataset
 * @return the index of the dataset, or -1 if it is not found
 */
int get_sds_index(const char *sds_name) {
  int i;

  for (i = 0 ; i < NUMBER_OF_SDS ; i++) {
    if (strcmp(sds_name, sds[i].name) == 0) return i;
  }

  return -1;
}

/** @brief parses options from command line 
 *
 * @param argc a pointer to the <b>argc</b> argument (argument count) from main
 * @param argv a pointer to the <b>*argv[]</b> argument (argument vector) from main
 * @return an option flag array stored in an <i>unsigned long</i> (a combination of <b>OPT_*</b> flags)
 */
unsigned long parse_options(int *argc, char **argv[]) {
  int option;
  int opt_flags;
  const char *options = "c:Ii:f:L:n:l:N:o:T:M:vx:w:Wz:";

  opt_flags = 0;
  opterr = 0;

  while ((option = getopt(*argc, *argv, options)) != -1) {
    switch (option) {
    case 'c' : {
      if (strcasecmp(optarg, "none") == 0) compression_type = COMP_CODE_NONE;
      else if (strcasecmp(optarg, "gzip") == 0) compression_type = COMP_CODE_DEFLATE;
      else if (strcasecmp(optarg, "huffman") == 0) compression_type = COMP_CODE_SKPHUFF;
      else {
	fprintf(stderr, APPNAME ": invalid value for compression type (option -c): %s (should be 'none', 'gzip' or 'huffman')\n", optarg);
	exit (EXIT_FAILURE);
      }
      break;
    }
    case 'W' : {
      opt_flags |= OPT_WIDTH_CORRECTION;
      break;
    }
    case 'I' : {
      opt_flags |= OPT_NO_VALIDITY_CHECK;
      break;
    }
    case 'i' : {
      incli = atof(optarg)*DEG2RAD;
      break;
    }
    case 'L' : {
      fprintf(stderr, APPNAME ": option -L currently not implemented\n");
      break;
    }
    case 'l' : {
      node_lon_in_degrees = atof(optarg);
      break;
    }
    case 'f' : {
      ilat_per_deg = atoi(optarg);
      if (ilat_per_deg <= 0) {
	fprintf(stderr, APPNAME ": invalid extraction factor: %d\n", ilat_per_deg);
	exit (EXIT_FAILURE);
      }
      break;
    }
    case 'o' : {
      strncpy(hdf_pathname, optarg, FILENAME_MAX - 1);
      break;
    }
    case 'T' : {
      Tsat = atof(optarg)*60; /* convert minutes into seconds */
      break;
    }
    case 'N' : {
      char *time_error;
      struct tm tm_node_time;
      time_error = strptime(optarg, time_format, &tm_node_time);
      if (time_error == NULL || *time_error != '\0') {
	fprintf(stderr, APPNAME ": option -N: %s: invalid time format (should be %s)\n", optarg, time_format);
	exit (EXIT_FAILURE);
      }
      node_time = custom_timegm(&tm_node_time);
      if (node_time == (time_t) -1) {
	fprintf(stderr, APPNAME ": option -N: invalid argument: %s\n", optarg);
	exit (EXIT_FAILURE);
      }
      break;
    }
    case 'M' : {
      char *time_error;
      struct tm tm_mean_time;
      time_error = strptime(optarg, time_format, &tm_mean_time);
      if (time_error == NULL || *time_error != '\0') {
	fprintf(stderr, APPNAME ": option -M: %s: invalid time format (should be %s)\n", optarg, time_format);
	exit (EXIT_FAILURE);
      }
      mean_time = custom_timegm(&tm_mean_time);
      if (mean_time == (time_t) -1) {
	fprintf(stderr, APPNAME ": option -M: invalid argument: %s\n", optarg);
	exit (EXIT_FAILURE);
      }
      break;
    }
    case 'v' : {
      opt_flags |= OPT_VERBOSE;
      break;
    }
    case 'x' : {
      char c = optarg[0];
      switch (c) {
      case 'c' :
	extract_mode = MSG_EXTRACT_CROSSTRACK;
	break;
      case 'h' :
	extract_mode = MSG_EXTRACT_HORIZONTAL;
	break;
      default :
	fprintf(stderr, APPNAME ": invalid argument for -x: should be c (crosstrack) or h (horizontal)\n");
	exit (EXIT_FAILURE);
	break;
      } // switch (c)
      break;
    }
    case 'w' : {
      half_width_in_degrees = atof(optarg)/2;
      break;
    }
    case 'z' : {
      compression_level = atoi(optarg);
      if (compression_level < 0 || compression_level > 9) {
	fprintf(stderr, APPNAME ": invalid value for compression level: %ld (should be between 0 and 9)\n", compression_level);
	exit (EXIT_FAILURE);
      }
      break;
    }
    case '?' : {
      break;
    }
    } /* switch */
  } /* while option */

  *argc -= optind - 1;
  *argv += optind - 1;
  return opt_flags;

} /* parse_options */

/** @brief main code
 *
 * Reads arguments from the command line.
 * Mandatory arguments are:
 * - an MSG/SEVIRI input file (format XRIT or FIS)
 * - a dataset name where the input data will be stored (the dataset belongs to the output file)
 *
 * If the output file doesn't exist, it will be created, otherwise only
 * the specified dataset will be overwritten (see <i>msg_extract_data</i> for details)
 *
 */

int main(int argc, char *argv[]) {
  
  unsigned long opt_flags = 0;
  char *input_pathname;
  msg_user_param_t user_param; /* structure used to pass user options and other data */
  size_t npix; /* number of pixels in the grid */
  int32 data_type; /* an HDF scalar data_type (should be DFNT_UINT8 or DFNT_UINT16) */
  xrit_rows_infos_t rows_infos; /* contains the rows validity and time infos (read from the input file) */
    
  hdf_pathname[0] = '\0';
  opt_flags = parse_options(&argc, &argv);

  geostat_init("msg");

  if (argc < 3) {
    usage();
    exit(EXIT_FAILURE);
  }

  input_pathname = argv[1];
  if (opt_flags & OPT_VERBOSE) printf(APPNAME ": input file: %s\n", input_pathname);

  sds_index = get_sds_index(argv[2]);
  if (sds_index < 2 || sds_index > NUMBER_OF_SDS - 1) { /* sds_indexes 0 and 1 are reserved (Latitude and Longitude) */
    fprintf(stderr, APPNAME ": unexpected sds name: %s\n", argv[2]);
    exit(EXIT_FAILURE);
  }

  if (hdf_pathname[0] == '\0') { /* no output name was specified by the user (see option -o in parse_options), so derive it from input_pathname */
    size_t len;
    char *input_filename = strrchr(input_pathname, PATHNAME_SEP); /* remove dirname from pathname, keep only part of the filename 
								   * following the last PATHNAME_SEP, ie '/' on Unix systems)
								   */
    
    if (input_filename == NULL) input_filename = input_pathname;
    else input_filename++ ; /* go to the first char after '/' (beginning of filename) */
 
    len = strlen(input_filename);
    strncpy(hdf_pathname, input_filename, FILENAME_MAX - 1);
    strncpy(&hdf_pathname[len], hdf_file_extension, FILENAME_MAX - 1 - len);
  }
  assert(hdf_pathname[0] != '\0');

  /* reads counts from input file (FIS or XRIT), read_any is a wrapper to read_fis and read_xrit.
   * counts is allocated by read_any
   */
  if (read_any(input_pathname, (void **) &counts, &npix, &data_type, &rows_infos, NUMBER_OF_SDS, sds, 1 /* logs enabled */) != 0) {
    exit(EXIT_FAILURE);
  }

  assert(npix == MSG_NLINES*MSG_NCOLUMNS);
  assert(counts != NULL);

  assert(half_width_in_degrees >= 0.);

  /* the following infos will be passed to msg_extract_data */
  user_param.half_width_in_radians = half_width_in_degrees*DEG2RAD;
  user_param.Tsat = Tsat;
  user_param.incli = incli;
  user_param.node_lon = node_lon_in_degrees*DEG2RAD;
  user_param.latmin = latmin;
  user_param.latmax = latmax;
  user_param.ilat_per_deg = ilat_per_deg;
  user_param.sds_index = sds_index;
  user_param.opt_flags = opt_flags;
  user_param.data_file = input_pathname;
  user_param.data_type = data_type;
  user_param.compression_type = compression_type;
  user_param.compression_level = compression_level;
  user_param.extract_mode = extract_mode;
  user_param.rows_infos = &rows_infos;
  user_param.node_time = node_time;
  user_param.mean_time = mean_time;

  if (user_param.half_width_in_radians > 0.) {
    strncpy(user_param.hdf_file, hdf_pathname, FILENAME_MAX - 1);
    msg_extract_data(extraction, counts, NUMBER_OF_SDS, sds, &user_param); /* the real work is done here */
  }

  exit (EXIT_SUCCESS);

}
