/* read_any.c */

#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <fnmatch.h> /* pattern matching */
#include "read_fis.h"
#include "read_any.h"
#include "msg_grid.h"

#define PATTERN_XRIT "H-*-MSG1__-MSG1________-*-CYCLE____-*"
#define PATTERN_FIS  "*.fis"

#undef DEBUG_READ_ANY

/* returns 1 (true) if 'filename' is an xrit file, 0 (false) otherwise */
static int is_xrit(const char *filename) {
  if (fnmatch(PATTERN_XRIT, filename, 0) != 0) return 0;
  return 1;
}

/* returns 1 (true) if 'filename' is a fis file, 0 (false) otherwise */
static int is_fis(const char *filename) {
  if (fnmatch(PATTERN_FIS, filename, 0) != 0) return 0;
  return 1;
}

/* loads 'buffer' from 'filename', and returns 'npix' (number of pixels) and 'data_type' (an HDF data type)
 * the array 'sds' (of size 'number_of_sds') may be updated by read_any (calibration scale and offset for each element)
 * error messages are sent to stderr if 'log' is not 0 (false)
 * returns 0 if no error occured
 *
 * Sample of use:
 *
 * unsigned void *buffer;
 * size_t npix;
 * int32 data_type;
 * xrit_rows_infos_t rows_infos;
 * sds_props_t sds[] = { array content };
 * size_t number_of_sds = sizeof(sds)/sizeof(*sds);
 * ret = read_any(filename, &buffer, &npix, &data_type, rows_infos, number_of_sds, sds, 0);
 */
int read_any(const char *filename, void **ppBuffer, size_t *pNpix, int32 *pData_type, xrit_rows_infos_t *rows_infos, int number_of_sds, sds_props_t *sds, const unsigned char log) {

#ifndef APPNAME
# define APPNAME "cookie-cutter"
#endif

  int ret = 0;
  char *shortname;
  size_t isds;

  assert(filename != NULL);
  shortname = strrchr(filename, '/'); /* remove dirname from filename */
  if (shortname == NULL) shortname = (char *) filename;
  else shortname++; /* points to the first char AFTER the last '/' */

  if (is_xrit(shortname)) {
    const size_t npix = MSG_NLINES*MSG_NCOLUMNS;
    MSG_Prologue prologue;

#ifdef DEBUG_READ_ANY
    fprintf(stderr, APPNAME ": read_any: %s: format xrit\n", filename);
#endif

    if (ppBuffer) {
      uint16_t *pBuffer;

      pBuffer  = calloc(npix, sizeof(uint16_t));
      assert(pBuffer != NULL);
      *((uint16_t **) ppBuffer) = pBuffer;
      xrit_orientation(XRIT_NORTH_AT_TOP);
      ret = read_xrit(filename, pBuffer, NULL, NULL, rows_infos, &prologue, NULL);
      if (ret != 10) {
	if (log) fprintf(stderr, APPNAME ": read_any: %s: 10 segments expected (1 prologue, 8 images segments and 1 epilogue), %d could be read\n", filename, ret);
	return -1;
      }
    }
    if (pNpix) *pNpix = npix;
    if (pData_type) *pData_type = DFNT_UINT16;
    if (sds) {
      for (isds = 0 ; isds < number_of_sds ; isds++) {
	int ichannel = xrit_met8_channel(sds[isds].name);
	double cal;
	double offset;
	if (ichannel < 0) continue;
	cal = ntoh_double(prologue.radiometricProcessing.level1_5ImageCalibration[ichannel].cal_Slope);
	offset = ntoh_double(prologue.radiometricProcessing.level1_5ImageCalibration[ichannel].cal_Offset);
	sds[isds].cal = cal;
	sds[isds].offset = -offset/cal; /* transformation from Seviri convention (a',b') 
					 * to HDF convention (a,b) for calibration : a(x-b) = a'x + b' */
      }
    } /* if (sds) */
  }
  else if (is_fis(shortname)) {
    size_t pix_size;
    size_t npix;
    int i;

#ifdef DEBUG_READ_ANY
    fprintf(stderr, APPNAME ": read_any: %s: format fis\n", filename);
#endif

    ret = read_fis(filename, ppBuffer, &npix, &pix_size, log);
    if (ret != 0) return -1;
    assert(npix == MSG_NLINES*MSG_NCOLUMNS);

    switch (pix_size) {
    case 1: { 
      if (pData_type) *pData_type = DFNT_UINT8; 
      break; 
    }
    case 2: { 
      if (pData_type) *pData_type = DFNT_UINT16; 
      break; 
    }
    case 4 : { 
      if (pData_type) *pData_type = DFNT_UINT32; 
      break; 
    }
    default: {
      fprintf(stderr, APPNAME ": read_any(%s): fatal error: unexpected value for pix_size: %d\n", filename, pix_size);
      abort();
      break;
    }
    } /* switch (pix_size) */

    if (pNpix) *pNpix = npix;

    for (i = 0 ; i < XRIT_NROWS ; i++) rows_infos->validity[i] = 1;

  }
  else {
    if (log) fprintf(stderr, APPNAME ": read_any: %s: unknown format: should be xrit or fis\n", filename);
    return -1;
  }

  return 0;
}
