/* seviri_radconv.c */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "seviri_read_raw.h"

#define XRIT_NROWS 3712
#define XRIT_NCOLS 3712
#define FILENAME_MAXLEN 256

typedef float xfloat_t; /* xfloat_t should be float or double */
const xfloat_t NaN = 0./0.;

static char output_file[FILENAME_MAXLEN] = "";
static xfloat_t input_buffer[XRIT_NROWS*XRIT_NCOLS];
static xfloat_t output_buffer[XRIT_NROWS*XRIT_NCOLS];

/* parses options from command line */
void parse_options(int *argc, char **argv[]) {
  int option;
  const char *options = "";

  opterr = 0;

  while ((option = getopt(*argc, *argv, options)) != -1) {
    switch (option) {
    case '?' : {
      break;
    }
    } /* switch */
  } /* while option */
  
  *argc -= optind - 1;
  *argv += optind - 1;
  
} /* parse_options */

int which_solar_band(const char *filename) {

  if (strstr(filename, "VIS006"))  return 0;
  else if (strstr(filename, "VIS008")) return 1;
  else if (strstr(filename, "IR016") || strstr(filename, "IR_016")) return 2;
  /* HRV not implemented for the time being, should return 3 */
  else return -1; /* default */
}

int which_thermic_band(const char *filename) {

  if (strstr(filename, "IR039") || strstr(filename, "IR_039")) return 0;
  else if (strstr(filename, "WV062") || strstr(filename, "WV_062")) return 1;
  else if (strstr(filename, "WV073") || strstr(filename, "WV_073")) return 2;
  else if (strstr(filename, "IR087") || strstr(filename, "IR_087")) return 3;
  else if (strstr(filename, "IR097") || strstr(filename, "IR_097")) return 4;
  else if (strstr(filename, "IR108") || strstr(filename, "IR_108")) return 5;
  else if (strstr(filename, "IR120") || strstr(filename, "IR_120")) return 6;
  else if (strstr(filename, "IR134") || strstr(filename, "IR_134")) return 7;
  else return -1; /* default */

}


void convert_rad_to_temp(const int band_idx, const xfloat_t *radiances, xfloat_t *temperatures) {

  /* All coefficients from Eumetsat official documentation */
  /* Band index :             0         1         2         3         4         5        6         7  */
  /*                          IR 3.9    WV6.2     WV7.3     IR8.7     IR9.7     IR10.8   IR12.0    IR13.4 */
  const xfloat_t vc[8] = {2569.094, 1598.566, 1362.142, 1149.083, 1034.345, 930.659, 839.661 , 752.381};
  const xfloat_t A[8] =  {0.9959  , 0.9963  , 0.9991  , 0.9996  , 0.9999  , 0.9983 , 0.9988  , 0.9981}; 
  const xfloat_t B[8] =  {3.471   , 2.219   , 0.485   , 0.181   , 0.060   , 0.627  , 0.397   , 0.576 }; 
  
  const xfloat_t c1 = 0.0000119104;
  const xfloat_t c2 = 1.43877;

  int irow, icol;

  for (irow = 0 ; irow < XRIT_NROWS ; irow++) {
    for (icol = 0 ; icol < XRIT_NCOLS ; icol++) {
      xfloat_t rad = radiances[irow*XRIT_NCOLS + icol];
      if (rad > 0) temperatures[irow*XRIT_NCOLS + icol] =  ( c2*vc[band_idx] / logf(c1 * vc[band_idx] * vc[band_idx] * vc[band_idx] / rad + 1.) - B[band_idx] )  / A[band_idx];
      else temperatures[irow*XRIT_NCOLS + icol] = NaN;
    }
  }
}

void convert_rad_to_refl(const int band_idx, const xfloat_t *radiances, xfloat_t *reflectances) {

  /* Band index :          0        1        2        3
                           VIS006   VIS008   IR016    HRV */
  const xfloat_t inv_Im[4] = { 1./65.2296, 1./73.0127, 1./62.3715, 1./78.8952 };
  static xfloat_t d2 = 1.; /* square of Sun-Earth distance (in AU) */

  int irow, icol;

  for (irow = 0 ; irow < XRIT_NROWS ; irow++) {
    for (icol = 0 ; icol < XRIT_NCOLS ; icol++) {
      xfloat_t rad = radiances[irow*XRIT_NCOLS + icol];
      if (rad > 0) reflectances[irow*XRIT_NCOLS + icol] = M_PI * rad * d2 * inv_Im[band_idx];
      else reflectances[irow*XRIT_NCOLS + icol] = NaN;
    }
  }
}

int main(int argc, char *argv[]) {

  FILE *input_stream = NULL;
  FILE *output_stream = NULL;
  char *input_file;
  char *input_shortname;
  size_t nread;
  size_t nwritten;
  int band_index;
  char *output_extension = "";
  void (*convert_rad)(const int, const xfloat_t *, xfloat_t *);

  parse_options(&argc, &argv);

  if (argc < 2) {
    fprintf(stderr, "usage: %s <seviri_radiances (raw file)> [<output_file (raw file)>]\n\n", argv[0]);
    fprintf(stderr, "raw file format must be: a 3712x3712 binary grid (in floating simple precision, 32 bits elements)\n\n");
    exit(EXIT_FAILURE);
  }

  input_file = argv[1];
  input_shortname = strrchr(input_file, '/');
  if (input_shortname) input_shortname++;
  else input_shortname = input_file;

  /* looks for a thermic band in the filename */
  band_index = which_thermic_band(input_shortname);
  if (band_index >= 0) {
    output_extension = "BrightTemp";
    convert_rad = convert_rad_to_temp;
    goto process;
  }
  
  /* if it failed, looks for a solar band */
  band_index = which_solar_band(input_shortname);
  if (band_index >= 0) {
    output_extension = "Reflect";
    convert_rad = convert_rad_to_refl;
    goto process;
  }

  fprintf(stderr, "%s: %s: a SEVIRI channel name was expected\n", argv[0], input_file);
  exit (EXIT_FAILURE);

 process:

  if (argc > 2) {
    strncpy(output_file, argv[2], FILENAME_MAXLEN - 1);
  }
  else {
    snprintf(output_file, FILENAME_MAXLEN - 1, "%s.%s", input_shortname, output_extension);
  }

  input_stream = fopen(input_file, "r");
  if (! input_stream) {
    perror(input_file);
    exit (EXIT_FAILURE);
  }

  nread = fread(input_buffer, sizeof(input_buffer), 1, input_stream);
  if (nread != 1) {
    perror(input_file);
    exit (EXIT_FAILURE);
  }

  if (fclose(input_stream) != 0) {
    perror(input_file);
    exit (EXIT_FAILURE);
  }

  convert_rad(band_index, input_buffer, output_buffer);
  
  output_stream = fopen(output_file, "w");
  if (! output_stream) {
    perror(output_file);
    exit (EXIT_FAILURE);
  }

  nwritten = fwrite(output_buffer, sizeof(output_buffer), 1, output_stream);
  if (nwritten != 1) {
    perror(output_file);
    exit (EXIT_FAILURE);
  }

  if (fclose(output_stream) != 0) {
    perror(output_file);
    exit (EXIT_FAILURE);
  }

  exit (EXIT_SUCCESS);

}
