/* seviri_zenith.c */

#include <stdio.h>
#include <stdlib.h>
#include "geostat/geostat.h"
#include "seviri_read_raw.h"
#include "math.h"

#define APPNAME "seviri_zenith_computation"
#define XRIT_NROWS 3712
#define XRIT_NCOLS 3712

typedef float xfloat_t; /* xfloat_t should be float or double */
static xfloat_t zenith_buffer[XRIT_NROWS*XRIT_NCOLS];
static xfloat_t latitude[XRIT_NROWS*XRIT_NCOLS];
static xfloat_t longitude[XRIT_NROWS*XRIT_NCOLS];

void seviri_compute_zenith(xfloat_t *zenith_buffer) {
  const int nrows = XRIT_NROWS;
  const int ncols = XRIT_NCOLS;
  double azimut;
  double zenith;
  int irow;
  int icol;
  geostat_err_t err;

#ifndef APPNAME
# define APPNAME "unknown"
#endif

  geostat_init("msg");

  for (irow = 0 ; irow < nrows ; irow++) {
    for (icol = 0 ; icol < ncols ; icol++) {
      float lat;
      float lon;
      
      lat = latitude[irow*ncols+icol]*M_PI/180.;
      if (lat < -M_PI_2 || lat > M_PI_2) continue;
      lon = longitude[irow*ncols+icol]*M_PI/180.;
      if (lon < -M_PI || lon > M_PI) continue;
      err = geostat_get_azimut_and_zenithal(lat, lon, &azimut, &zenith);
      if (err != GEOSTAT_ERR_OK) {
	switch (err) {
	case (GEOSTAT_ERR_OUT_OF_RANGE) :
	  fprintf(stderr, APPNAME ": seviri_compute_zenith: GEOSTAT_ERR_OUT_OF_RANGE\n");
	  exit (EXIT_FAILURE);
	case (GEOSTAT_ERR_BAD_INIT) :
	  fprintf(stderr, APPNAME ": seviri_compute_zenith: GEOSTAT_ERR_BAD_INIT\n");
	  exit (EXIT_FAILURE);
	case (GEOSTAT_ERR_BAD_INPUT_COORD) :
	  fprintf(stderr, APPNAME ": seviri_compute_zenith: GEOSTAT_ERR_BAD_INPUT_COORD\n");
	  exit (EXIT_FAILURE);
	default :
	  fprintf(stderr, APPNAME ": seviri_compute_zenith: unexpected value in variable err\n");
	  exit (EXIT_FAILURE);
	} /* switch (err) */
      }
      zenith_buffer[irow*ncols + icol] = (xfloat_t) cos(zenith);
    }
  }
}

int main(int argc, char *argv[]) {

  char *latitude_file = "seviri_latitudes.raw";
  char *longitude_file = "seviri_longitudes.raw";
  char *zenith_file = "seviri_zenith.raw";
  FILE *zenith_stream;

  if (argc < 2) {
    fprintf(stderr, "usage: %s </path/to/seviri_latitudes.raw> </path/to/seviri_longitudes.raw>\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  latitude_file = argv[1];
  longitude_file = argv[2];

  seviri_read_raw(latitude_file, latitude);
  seviri_read_raw(longitude_file, longitude);

  zenith_stream = fopen(zenith_file, "w");
  if (zenith_stream == NULL) {
    perror(zenith_file);
    exit(EXIT_FAILURE);
  }

  seviri_compute_zenith(zenith_buffer);
  fwrite(zenith_buffer, XRIT_NROWS*XRIT_NCOLS*sizeof(xfloat_t), 1, zenith_stream);

  if (fclose(zenith_stream) != 0) {
    perror(zenith_file);
    exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS);

}
