/* xrit2raw.c */

/* 
   Copyright (C) 2005 Fabrice Ducos <fabrice.ducos@icare.univ-lille1.fr>
   This file is part of the LibXRIT Library.

   The LibXRIT Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The LibXRIT Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the LibXRIT Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include "xrit_swap.h"
#include "libxrit.h"

#define FILENAME_MAXLEN 256

#define OPT_PROLOGUE 1
#define OPT_UINT16 1
#define OPT_FLOAT  2
#define OPT_DOUBLE 3

static unsigned char spare[19814];

static uint16_t counts[XRIT_NROWS][XRIT_NCOLUMNS];
static float  spectral_radiances_flt[XRIT_NROWS][XRIT_NCOLUMNS];
static double spectral_radiances_dbl[XRIT_NROWS][XRIT_NCOLUMNS];

char prologue_file[FILENAME_MAXLEN] = "";

int data_type = OPT_UINT16;
int isegment = 0;

/* parses options from command line */
void parse_options(int *argc, char **argv[]) {
  int option;
  const char *options = "P:n:rsfd";

  opterr = 0;

  while ((option = getopt(*argc, *argv, options)) != -1) {
    switch (option) {
    case 'P' : {
      strncpy(prologue_file, optarg, FILENAME_MAXLEN - 1);
      break;
    }
    case 'n' : {
      isegment = atoi(optarg);
      if (isegment < 1 || isegment > 8) {
	fprintf(stderr, "xrit2raw: invalid argument for -n (should be between 1 and 8), will be ignored\n");
	isegment = 0;
      }
      break;
    }
    case 'r' : {
      xrit_orientation(XRIT_SOUTH_AT_TOP);
      break;
    }
    case 's' : {
      data_type = OPT_UINT16;
      break;
    }
    case 'f' : {
      data_type = OPT_FLOAT;
      break;
    }
    case 'd' : {
      data_type = OPT_DOUBLE;
      break;
    }
    case '?' : {
      break;
    }
    } /* switch */
  } /* while option */
  
  *argc -= optind - 1;
  *argv += optind - 1;
  
} /* parse_options */



int main(int argc, char *argv[]) {

  FILE *output_stream;
  char *input_file;
  char output_file[FILENAME_MAXLEN];
  size_t nwritten;
  MSG_Prologue prologue;
  double cal_slope[12];
  double cal_offset[12];
  const unsigned char version = 0;
  int ichannel;
  register int irow, icol;
  int nsegments = 0;
  int i;

  xrit_orientation(XRIT_NORTH_AT_TOP);
  parse_options(&argc, &argv);

  if (argc < 2) {
    fprintf(stderr, "usage: %s [OPTIONS] <input xrit file> [<output raw file>]\n\n", argv[0]);
    fprintf(stderr, 
	    "OPTIONS:\n"
	    "  -P <prologue_file>                 saves the msg prologue into 'prologue_file'\n"
	    "  -n <segment>                       extracts only one segment (between 1 and 8), over 464 rows\n"
	    "  -r                                 the image will be reversed\n"
	    "  -s                                 stores msg counts (short integer) into the raw file (default behavior)\n"
	    "  -f                                 stores msg calibrated data as float into the raw file\n"
	    "  -d                                 stores msg calibrated data as double into the raw file\n");
    exit(EXIT_FAILURE);
  }

  input_file = argv[1];

  char *input_shortname = strrchr(input_file, '/');
  if (input_shortname) input_shortname++;
  else input_shortname = input_file;

  ichannel = xrit_met8_channel(input_shortname);

  if (argc > 2) {
    strncpy(output_file, argv[2], FILENAME_MAXLEN - 1);
  }
  else {
    snprintf(output_file, FILENAME_MAXLEN - 1, "%s.raw", input_shortname);
  }

  if ((nsegments = read_xrit(input_file, counts[0], &isegment, NULL, NULL, &prologue, NULL)) < 0) {
    exit (EXIT_FAILURE);
  }

  if (*prologue_file) { /* writes prologue into a file */
    FILE *prologue_stream;
    prologue_stream = fopen(prologue_file, "w");
    if (! prologue_stream) perror(prologue_file);
    else {
      nwritten = fwrite(&version, sizeof(version), 1, prologue_stream);
      if (nwritten != 1) {
	perror(prologue_file);
	exit(EXIT_FAILURE);
      }
      nwritten = fwrite(&prologue, sizeof(prologue), 1, prologue_stream);
      if (nwritten != 1) {
	perror(prologue_file);
	exit(EXIT_FAILURE);
      }
      /* fills 0 in place of IMPFConfiguration record (currently not implemented in prologue) */
      nwritten = fwrite(spare, sizeof(spare), 1, prologue_stream);
      if (nwritten != 1) {
	perror(prologue_file);
	exit(EXIT_FAILURE);
      }
      
      if (fclose(prologue_stream) != 0) {
	perror(output_file);
	exit(EXIT_FAILURE);
      }
    }
  }


  for (i = 0 ; i < 12 ; i++) {
    cal_slope[i] = ntoh_double(prologue.radiometricProcessing.level1_5ImageCalibration[i].cal_Slope);
    cal_offset[i] = ntoh_double(prologue.radiometricProcessing.level1_5ImageCalibration[i].cal_Offset);
  }

  output_stream = fopen(output_file, "w");
  if (! output_stream) {
    perror(output_file);
    exit(EXIT_FAILURE);
  }

  switch (data_type) {
  case OPT_UINT16 :
    if (nsegments > 1) nwritten = fwrite(counts[0], sizeof(counts), 1, output_stream);
    else  nwritten = fwrite(counts[0], sizeof(counts)/8, 1, output_stream);
    break;
  case OPT_FLOAT :
    for (irow = 0 ; irow < XRIT_NROWS ; irow++) {
      for (icol = 0 ; icol < XRIT_NCOLUMNS ; icol++) {
	spectral_radiances_flt[irow][icol] = cal_slope[ichannel]*counts[irow][icol] + cal_offset[ichannel];
      }
    }
    if (nsegments > 1) nwritten = fwrite(spectral_radiances_flt[0], sizeof(spectral_radiances_flt), 1, output_stream);
    else nwritten = fwrite(spectral_radiances_flt[0], sizeof(spectral_radiances_flt)/8, 1, output_stream);
    break;
  case OPT_DOUBLE :
    for (irow = 0 ; irow < XRIT_NROWS ; irow++) {
      for (icol = 0 ; icol < XRIT_NCOLUMNS ; icol++) {
	spectral_radiances_dbl[irow][icol] = cal_slope[ichannel]*counts[irow][icol] + cal_offset[ichannel];
      }
    }
    if (nsegments > 1) nwritten = fwrite(spectral_radiances_dbl[0], sizeof(spectral_radiances_dbl), 1, output_stream);
    else nwritten = fwrite(spectral_radiances_dbl[0], sizeof(spectral_radiances_dbl)/8, 1, output_stream);
    break;
  default :
    fprintf(stderr, "%s: fatal error: unexpected value for variable data_type: %d\n", argv[0], data_type);
    exit (EXIT_FAILURE);
    break;
  }
   
  if (nwritten != 1) {
    perror(output_file);
    exit(EXIT_FAILURE);
  }

  if (fclose(output_stream) != 0) {
    perror(output_file);
    exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS);

}
